# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from argparse import ArgumentParser
from pathlib import Path

from discopop_library.PathManagement.PathManagement import get_path, get_path_or_none
from .discopop_explorer import ExplorerArguments, run


def parse_args() -> ExplorerArguments:
    """Parse the arguments passed to the discopop_explorer"""
    parser = ArgumentParser(description="DiscoPoP Explorer")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments for the task pattern detector and other experimental features. These flags are considered EXPERIMENTAL and they may or may not be removed or changed in the near future.",
    )

    # fmt: off
    parser.add_argument(
        "--path", type=str, default="./",
        help="Path to the .discopop directory to be analyzed")
    parser.add_argument(
        "--cu-xml", type=str, default="profiler/Data.xml",
        help="CU node xml file")
    parser.add_argument(
        "--dep-file", type=str, default="profiler/dynamic_dependencies.txt",
        help="Dependencies text file"
    )
    parser.add_argument(
        "--loop-counter", type=str, default="profiler/loop_counter_output.txt",
        help="Loop counter data"
    )
    parser.add_argument(
        "--reduction", type=str, default="profiler/reduction.txt",
        help="Reduction variables file"
    )
    parser.add_argument(
        "--fmap", type=str, default="FileMapping.txt",
        help="File mapping")
    parser.add_argument(
        "--plugins", type=str, nargs="*", default=[],
        help="Plugins to execute")
    parser.add_argument(
        "--dp-build-path", type=str, default=Path(__file__).resolve().parent.parent / "build",
        help="Path to DiscoPoP build folder"
    )
    # flags related to output and formatting:
    parser.add_argument(
        "--json", type=str, nargs="?", default="explorer/patterns.json",
        help="Json output")
    parser.add_argument(
        "--profiling", type=str, nargs="?", default=None, const="explorer/profiling_stats.txt",
        help="Enable profiling. If a path is given, the profiling stats are written to the given file, otherwise to profiling_stats.txt",
    )
    parser.add_argument(
        "--dump-pet", type=str, nargs="?", default=None, const="explorer/pet_dump.json",
        help="Dump PET Graph to JSON file. If a path is given, the PET Graph is written to the given file, otherwise to pet_dump.json",
    )
    parser.add_argument(
        "--dump-detection-result", type=str, nargs="?", default="explorer/detection_result_dump.json",
        help="Dump DetectionResult object to JSON file. If a path is given, the DetectionResult object is written to the given file, otherwise to detection_result_dump.json. Contents are equivalent to the json output. NOTE: This dump contains a dump of the PET Graph!",
    )
    parser.add_argument(
        "--enable-patterns", type=str, nargs="?", default="*",
        help="Specify comma-separated list of pattern types to be identified. Options: reduction,doall,pipeline,geodec,simplegpu. Default: *",
    )

    # EXPERIMENTAL FLAGS:
    # temporary flag for microbenchmark file
    experimental_parser.add_argument(
        "--microbench-file", type=str, default=None,
        help="path to a microbenchmark json file",
    )

    # flags related to task pattern detector:
    experimental_parser.add_argument(
        "--task-pattern", action="store_true",
        help="Enables the task parallelism pattern identification. Requires --cu-inst-res and --llvm-cxxfilt-path to be set.",
    )
#    experimental_parser.add_argument(
#        "--detect-scheduling-clauses", action="store_true",
#        help="Enables the detection of scheduling clauses for parallel loops.")

    experimental_parser.add_argument(
        "--generate-data-cu-inst", type=str, default=None,
        help="Generates Data_CUInst.txt file and stores it in the given directory. Stops the regular execution of the discopop_explorer. Requires --cu-xml, --dep-file, --loop-counter, --reduction.",
    )
    experimental_parser.add_argument(
        "--cu-inst-res", type=str, default=None,
        help="CU instantiation result file."
    )
    experimental_parser.add_argument(
        "--llvm-cxxfilt-path", type=str, default=None,
        help="Path to llvm-cxxfilt executable. Required for task pattern detector if non-standard path should be used.",
    )
    # fmt: on

    arguments = parser.parse_args()

    # ensure that --cu-inst-res and --llvm-cxxfilt-path are set if --task-pattern is set
    if arguments.task_pattern and (arguments.cu_inst_res is None or arguments.llvm_cxxfilt_path is None):
        parser.error("--task-pattern requires --cu-inst-res and --llvm-cxxfilt-path to be set")

    # ensure that --cu-xml, --dep-file, --loop-counter, --reduction are set if --generate-data-cu-inst is set
    # NOTE: no need to check, the defaults got us covered
    # if arguments.generate_data_cu_inst is not None and (arguments.cu_xml is None or arguments.dep_file is None or arguments.loop_counter is None or arguments.reduction is None):
    #    parser.error("--generate-data-cu-inst requires --cu-xml, --dep-file, --loop-counter, --reduction to be set")

    # post processing of arguments: interpret relative paths as relative to project path
    arguments.path = str(Path(arguments.path).resolve())
    arguments.cu_xml = get_path(arguments.path, arguments.cu_xml)
    arguments.dep_file = get_path(arguments.path, arguments.dep_file)
    arguments.loop_counter = get_path(arguments.path, arguments.loop_counter)
    arguments.reduction = get_path(arguments.path, arguments.reduction)
    arguments.fmap = get_path(arguments.path, arguments.fmap)
    arguments.json = get_path_or_none(arguments.path, arguments.json)
    arguments.cu_inst_res = get_path_or_none(arguments.path, arguments.cu_inst_res)
    arguments.generate_data_cu_inst = get_path_or_none(arguments.path, arguments.generate_data_cu_inst)
    arguments.profiling = get_path_or_none(arguments.path, arguments.profiling)
    arguments.dump_pet = get_path_or_none(arguments.path, arguments.dump_pet)
    arguments.dump_detection_result = get_path_or_none(arguments.path, arguments.dump_detection_result)
    arguments.microbench_file = get_path_or_none(arguments.path, arguments.microbench_file)

    return ExplorerArguments(
        discopop_build_path=arguments.dp_build_path,
        microbench_file=arguments.microbench_file,
        project_path=arguments.path,
        cu_xml_file=arguments.cu_xml,
        dep_file=arguments.dep_file,
        loop_counter_file=arguments.loop_counter,
        reduction_file=arguments.reduction,
        file_mapping_file=arguments.fmap,
        plugins=arguments.plugins,
        enable_task_pattern=arguments.task_pattern,
        #        detect_scheduling_clauses=arguments.detect_scheduling_clauses,
        detect_scheduling_clauses=False,
        enable_profiling_dump_file=arguments.profiling,
        enable_pet_dump_file=arguments.dump_pet,
        enable_detection_result_dump_file=arguments.dump_detection_result,
        enable_patterns=arguments.enable_patterns,
        generate_data_cu_inst=arguments.generate_data_cu_inst,
        cu_inst_result_file=arguments.cu_inst_res,
        llvm_cxxfilt_path=arguments.llvm_cxxfilt_path,
        enable_json_file=arguments.json,
    )


def main():
    arguments = parse_args()
    run(arguments)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
import sys
from typing import Set, Tuple, Dict, List, cast, Optional, Union

from networkx import MultiDiGraph  # type: ignore

from discopop_explorer.PEGraphX import (
    PEGraphX,
    EdgeType,
    NodeID,
    MemoryRegion,
    DepType,
    CUNode,
    FunctionNode,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.EntryPoint import EntryPoint
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import (
    UpdateType,
    EntryPointType,
    ExitPointType,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.ExitPoint import ExitPoint
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Update import Update


def convert_updates_to_entry_and_exit_points(
    pet: PEGraphX,
    issued_updates: Set[Update],
    memory_region_liveness_by_device: Dict[int, Dict[MemoryRegion, List[NodeID]]],
) -> Tuple[Set[EntryPoint], Set[ExitPoint], Set[Update]]:
    entry_points: Set[EntryPoint] = set()
    exit_points: Set[ExitPoint] = set()
    updates: Set[Update] = set()

    for issued_update in issued_updates:
        if issued_update.update_type == UpdateType.TO_DEVICE or issued_update.update_type == UpdateType.ALLOCATE:
            # check if the memory region was live on the device before the update
            qualifies_as_entry = True
            for mem_reg in issued_update.memory_regions:
                if mem_reg not in memory_region_liveness_by_device[1]:
                    continue
                if issued_update.synchronous_source_cu_id in memory_region_liveness_by_device[1][mem_reg]:
                    qualifies_as_entry = False
                    break
            if qualifies_as_entry:
                if issued_update.update_type == UpdateType.TO_DEVICE:
                    if issued_update.asynchronous_possible:
                        # asynchronous entry possible
                        enp = EntryPoint(
                            pet,
                            issued_update.variable_names,
                            issued_update.memory_regions,
                            cast(NodeID, issued_update.asynchronous_source_cu_id),
                            issued_update.sink_cu_id,
                            EntryPointType.ASYNC_TO_DEVICE,
                        )
                        # add dependencies to EntryPoint
                        for dependency in issued_update.dependencies:
                            enp.dependencies.add(dependency)
                        entry_points.add(enp)
                    else:
                        # asynchronous entry not possible
                        # create a synchronous entry point
                        entry_points.add(
                            EntryPoint(
                                pet,
                                issued_update.variable_names,
                                issued_update.memory_regions,
                                issued_update.synchronous_source_cu_id,
                                issued_update.sink_cu_id,
                                EntryPointType.TO_DEVICE,
                            )
                        )
                else:
                    entry_points.add(
                        EntryPoint(
                            pet,
                            issued_update.variable_names,
                            issued_update.memory_regions,
                            issued_update.synchronous_source_cu_id,
                            issued_update.sink_cu_id,
                            EntryPointType.ALLOCATE,
                        )
                    )
            else:
                updates.add(issued_update)

        elif issued_update.update_type == UpdateType.FROM_DEVICE:
            # check if the memory region is live on the device after the update
            qualifies_as_exit = True
            for mem_reg in issued_update.memory_regions:
                if mem_reg not in memory_region_liveness_by_device[1]:
                    continue
                if issued_update.sink_cu_id in memory_region_liveness_by_device[1][mem_reg]:
                    qualifies_as_exit = False
                    break
            if qualifies_as_exit:
                # check if asynchronous exiting is possible
                if issued_update.asynchronous_possible:
                    # asynchronous exit possible
                    exp = ExitPoint(
                        pet,
                        issued_update.variable_names,
                        issued_update.memory_regions,
                        cast(NodeID, issued_update.asynchronous_source_cu_id),
                        issued_update.sink_cu_id,
                        ExitPointType.ASYNC_FROM_DEVICE,
                    )
                    # add dependencies to EntryPoint
                    for dependency in issued_update.dependencies:
                        exp.dependencies.add(dependency)
                    exit_points.add(exp)
                else:
                    # asynchronous exiting not possible
                    exit_points.add(
                        ExitPoint(
                            pet,
                            issued_update.variable_names,
                            issued_update.memory_regions,
                            issued_update.synchronous_source_cu_id,
                            issued_update.sink_cu_id,
                            ExitPointType.FROM_DEVICE,
                        )
                    )
            else:
                updates.add(issued_update)
        else:
            raise ValueError("Unsupported update type: ", issued_update.update_type)

    return entry_points, exit_points, updates


def add_aliases(
    pet: PEGraphX,
    issued_updates: Set[Update],
    memory_regions_to_functions_and_variables: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]],
) -> Set[Update]:
    for update in issued_updates:
        source_parent_function_node = pet.get_parent_function(pet.node_at(update.synchronous_source_cu_id))
        sink_parent_function_node = pet.get_parent_function(pet.node_at(update.sink_cu_id))
        if source_parent_function_node == sink_parent_function_node:
            # add alias information from function level
            modification_found = True
            while modification_found:
                modification_found = False
                # add missing variable names
                for mem_reg in update.memory_regions:
                    alias_var_names = memory_regions_to_functions_and_variables[mem_reg][source_parent_function_node.id]
                    len_pre = len(update.variable_names)
                    update.variable_names.update(alias_var_names)
                    # add variable name to dependency if required
                    for dependency in update.dependencies:
                        if mem_reg in dependency.memory_regions:
                            dependency.var_names.update(alias_var_names)
                    len_post = len(update.variable_names)
                    if len_pre != len_post:
                        modification_found = True

                # add missing memory regions
                for var_name in update.variable_names:
                    for mem_reg in memory_regions_to_functions_and_variables:
                        if source_parent_function_node.id not in memory_regions_to_functions_and_variables[mem_reg]:
                            continue
                        potential_aliases = memory_regions_to_functions_and_variables[mem_reg][
                            source_parent_function_node.id
                        ]
                        if var_name in potential_aliases:
                            len_pre = len(update.memory_regions)
                            update.memory_regions.add(mem_reg)
                            len_post = len(update.memory_regions)
                            if len_pre != len_post:
                                modification_found = True

    return issued_updates


def identify_end_of_life_points(
    pet: PEGraphX,
    entry_points: Set[EntryPoint],
    exit_points: Set[ExitPoint],
    memory_region_liveness_by_device: Dict[int, Dict[MemoryRegion, List[NodeID]]],
    memory_regions_to_functions_and_variables: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]],
) -> Set[ExitPoint]:
    eol_points: Set[Tuple[NodeID, NodeID, Tuple[MemoryRegion, ...]]] = set()

    print("Identifying end of live for variables...", file=sys.stderr)

    # get list of seen aliases:
    mem_reg_aliases: Dict[MemoryRegion, Set[MemoryRegion]] = dict()
    for entry_point in entry_points:
        for mem_reg in entry_point.memory_regions:
            if mem_reg not in mem_reg_aliases:
                mem_reg_aliases[mem_reg] = set()
            mem_reg_aliases[mem_reg].update(entry_point.memory_regions)
    for exit_point in exit_points:
        for mem_reg in exit_point.memory_regions:
            if mem_reg not in mem_reg_aliases:
                mem_reg_aliases[mem_reg] = set()
            mem_reg_aliases[mem_reg].update(exit_point.memory_regions)

    print("\tUsed Aliases: ", mem_reg_aliases, file=sys.stderr)

    # actual search for EOL
    for device_id in memory_region_liveness_by_device:
        if device_id == 0:
            # do not search for EOL on host
            continue
        for mem_reg in memory_region_liveness_by_device[device_id]:
            # check if mem_reg is live in all successors of all contained cu's
            for cu_id in memory_region_liveness_by_device[device_id][mem_reg]:
                for successor_node in pet.direct_successors(pet.node_at(cu_id)):
                    if successor_node.id not in memory_region_liveness_by_device[device_id][mem_reg]:
                        # mem_reg is not live anymore. create an EOL point
                        if mem_reg in mem_reg_aliases:
                            eol_points.add(
                                (
                                    cu_id,
                                    successor_node.id,
                                    tuple(mem_reg_aliases[mem_reg]),
                                )
                            )
                        else:
                            eol_points.add((cu_id, successor_node.id, tuple([mem_reg])))

    # remove eols which are covered by known exit points
    for exit_point in exit_points:
        to_be_removed: Set[Tuple[NodeID, NodeID, Tuple[MemoryRegion, ...]]] = set()
        for eol in eol_points:
            # check if either exit_point is reachable from eol or vice versa.
            # in both cases, eol is covered by the exit_point and can be ignored
            if (
                (
                    pet.is_predecessor(exit_point.source_cu_id, eol[0])
                    and pet.is_predecessor(exit_point.sink_cu_id, eol[1])
                )
                or pet.is_predecessor(eol[0], exit_point.source_cu_id)
                and pet.is_predecessor(eol[1], exit_point.sink_cu_id)
            ):
                # check if memory regions overlap, i.e. if the exit_point covers eol
                if len([mem_reg for mem_reg in exit_point.memory_regions if mem_reg in eol[2]]) > 0:
                    # overlap exists, remove eol
                    to_be_removed.add(eol)

        for elem in to_be_removed:
            if elem in eol_points:
                eol_points.remove(elem)

    eol_exit_points: Set[ExitPoint] = set()
    for eol in eol_points:
        parent_function_node_id = pet.get_parent_function(pet.node_at(eol[0])).id
        var_names: Set[VarName] = set()
        for mem_reg in eol[2]:
            if parent_function_node_id in memory_regions_to_functions_and_variables[mem_reg]:
                var_names.update(memory_regions_to_functions_and_variables[mem_reg][parent_function_node_id])
        memory_regions = set(eol[2])
        # check if the exited data is required by another function
        # if so, mark the exit point as ExitPointType.FROM
        path_nodes = pet.get_path_nodes_between(
            cast(CUNode, pet.node_at(eol[1])),
            cast(CUNode, pet.node_at(eol[0])),
            [EdgeType.SUCCESSOR, EdgeType.CHILD],
        )

        in_raw_edges_from_outside = []
        for path_node in path_nodes:
            in_raw_edges_from_outside += [
                (s, t, d)
                for s, t, d in pet.in_edges(path_node.id, EdgeType.DATA)
                if d.dtype == DepType.RAW
                and d.memory_region in memory_regions
                and pet.get_parent_function(pet.node_at(s)) != pet.get_parent_function(pet.node_at(t))
            ]
        if len(in_raw_edges_from_outside) > 0:
            # value is read -> Copy back to the host so the value does not get discarded
            eol_exit_points.add(ExitPoint(pet, var_names, memory_regions, eol[0], eol[1], ExitPointType.FROM_DEVICE))
        else:
            # otherwise, mark it as ExitPointType.DELETE
            eol_exit_points.add(ExitPoint(pet, var_names, memory_regions, eol[0], eol[1], ExitPointType.DELETE))

    print("\tDone.", file=sys.stderr)
    print(file=sys.stderr)

    return eol_exit_points


def extend_region_liveness_using_unrolled_functions(
    pet: PEGraphX,
    liveness: Dict[MemoryRegion, List[NodeID]],
    unrolled_function_graphs: Dict[FunctionNode, MultiDiGraph],
) -> Dict[MemoryRegion, List[NodeID]]:
    # TODO: potential optimization: invert the 'liveness' dict to allow faster membership check

    for function in pet.all_nodes(FunctionNode):
        print("FUNCTION: ", function.name)
        unrolled_function_graph = unrolled_function_graphs[function]
        queue: List[Tuple[NodeID, MemoryRegion, List[NodeID]]] = []
        # initialize queue
        # entry_node_id = function.get_entry_cu_id(pet)
        for child_node_id in cast(List[NodeID], function.children_cu_ids):
            for mem_reg in liveness:
                if child_node_id in liveness[mem_reg]:
                    queue.append((child_node_id, mem_reg, []))

        # process queue
        while queue:
            print("Queue len: ", len(queue))
            current_node_id, current_mem_reg, visited_nodes = queue.pop()
            visited_nodes.append(current_node_id)

            successors = [t for s, t, d in unrolled_function_graph.out_edges(current_node_id, data="data")]

            for successor in successors:
                # if mem_reg is live in successor, create a new, clean queue entry and set all visited nodes to live
                # else, create a new queue entry for the successor and use visited_nodes again

                if successor in liveness[current_mem_reg]:
                    # create a clean queue entry
                    queue.append((successor, current_mem_reg, []))
                    # set visited nodes to live
                    for node_id in visited_nodes:
                        if node_id not in liveness[current_mem_reg]:
                            liveness[current_mem_reg].append(node_id)
                            print("Set ", current_mem_reg, " live in ", node_id)
                else:
                    queue.append((successor, current_mem_reg, copy.deepcopy(visited_nodes)))

    return liveness


def remove_duplicates(target_set: Union[Set[Update], Set[EntryPoint], Set[ExitPoint]]):
    to_be_removed = []
    for element_1 in target_set:
        for element_2 in target_set:
            if element_1 is element_2:
                continue
            if element_1 == element_2:
                to_be_removed.append(element_2)

    for element in to_be_removed:
        if element in target_set:
            target_set.remove(element)  # type: ignore

    return target_set


def join_elements(target_set):
    grouping_dict = dict()

    for elem in target_set:
        if elem.get_position_identifier() not in grouping_dict:
            grouping_dict[elem.get_position_identifier()] = []
        grouping_dict[elem.get_position_identifier()].append(elem)

    result_set = set()

    for key in grouping_dict:
        # assemble a single EntryPoint
        joined_entry_point: Optional[EntryPoint] = None
        for elem in grouping_dict[key]:
            if joined_entry_point is None:
                joined_entry_point = elem
            else:
                joined_entry_point.join(elem)
        result_set.add(cast(EntryPoint, joined_entry_point))
    return result_set



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Dict, Set

from discopop_explorer.PEGraphX import PEGraphX, NodeID, MemoryRegion
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)


def propagate_variable_name_associations(
    pet: PEGraphX,
    memory_regions_to_cus_and_variables: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]],
) -> Dict[MemoryRegion, Dict[NodeID, Set[VarName]]]:
    """Replaces individual cu ids with id's of their parent functions."""
    updated_dict: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]] = dict()

    for mem_reg in memory_regions_to_cus_and_variables:
        for cu_id in memory_regions_to_cus_and_variables[mem_reg]:
            var_names = memory_regions_to_cus_and_variables[mem_reg][cu_id]
            cu_node = pet.node_at(cu_id)
            # get parent function
            parent = pet.get_parent_function(cu_node)

            # save variable name association for parent function
            if mem_reg not in updated_dict:
                updated_dict[mem_reg] = dict()
            if parent.id not in updated_dict[mem_reg]:
                updated_dict[mem_reg][parent.id] = set()
            updated_dict[mem_reg][parent.id].update(var_names)

    return updated_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.classes.context.Update import Update
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


def remove_duplicated_updates(configuration: OptimizerOutputPattern, arguments: OptimizerArguments):
    cleaned_updates: List[Update] = []
    buffer: List[str] = []

    def get_buffer_str(updt, ignore_first_update=False):
        if ignore_first_update:
            result_str = ""
        else:
            result_str = "F" if updt.is_first_data_occurrence else ""
        return (
            result_str
            + "U("
            + str(updt.source_node_id)
            + "@"
            + str(updt.source_device_id)
            + ","
            + str(updt.target_node_id)
            + "@"
            + str(updt.target_device_id)
            + ","
            + str(updt.write_data_access.var_name)
            + ")"
        )

    # remove general duplicates
    for update in configuration.data_movement:
        if get_buffer_str(update) not in buffer:
            cleaned_updates.append(update)
            buffer.append(get_buffer_str(update))
        else:
            # duplicate found. Ignore.
            pass
    configuration.data_movement = cleaned_updates

    # remove update <-> first update duplicates
    buffer = []
    cleaned_updates = []
    regular_updates = [u for u in configuration.data_movement if not u.is_first_data_occurrence]
    first_updates = [u for u in configuration.data_movement if u.is_first_data_occurrence]
    # add first updates to cleaned list
    for update in first_updates:
        if get_buffer_str(update, ignore_first_update=True) not in buffer:
            cleaned_updates.append(update)
            buffer.append(get_buffer_str(update, ignore_first_update=True))
        else:
            # duplicate found. Ignore
            pass
    # check for duplicates in regular updates
    for update in regular_updates:
        if get_buffer_str(update, ignore_first_update=True) not in buffer:
            cleaned_updates.append(update)
            buffer.append(get_buffer_str(update, ignore_first_update=True))
        else:
            # duplicate found. Ignore
            pass
    configuration.data_movement = cleaned_updates

    if arguments.verbose:
        print("Removed duplicates from updates")
        for update in configuration.data_movement:
            print("# ", update)
        print()

    return configuration



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


from typing import List, Optional, cast

from discopop_explorer.PEGraphX import DummyNode, PEGraphX, MWType
from discopop_explorer.parser import parse_inputs
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.do_all_detector import run_detection as detect_do_all
from discopop_explorer.pattern_detectors.reduction_detector import run_detection as detect_reduction
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.filter import (
    filter_data_sharing_clauses,
    remove_useless_barrier_suggestions,
    remove_duplicate_data_sharing_clauses,
    filter_data_depend_clauses,
    remove_duplicates,
)
from discopop_explorer.pattern_detectors.task_parallelism.postprocessor import (
    group_task_suggestions,
    sort_output,
)
from discopop_explorer.pattern_detectors.task_parallelism.preprocessor import (
    cu_xml_preprocessing,
    check_loop_scopes,
)
from discopop_explorer.pattern_detectors.task_parallelism.suggesters.auxiliary import (
    suggest_parallel_regions,
    set_task_contained_lines,
    detect_taskloop_reduction,
    combine_omittable_cus,
)
from discopop_explorer.pattern_detectors.task_parallelism.suggesters.barriers import (
    detect_barrier_suggestions,
    suggest_barriers_for_uncovered_tasks_before_return,
    validate_barriers,
    suggest_missing_barriers_for_global_vars,
)
from discopop_explorer.pattern_detectors.task_parallelism.suggesters.data_sharing_clauses import (
    suggest_shared_clauses_for_all_tasks_in_function_body,
)
from discopop_explorer.pattern_detectors.task_parallelism.suggesters.dependency_clauses import (
    detect_dependency_clauses_alias_based,
)
from discopop_explorer.pattern_detectors.task_parallelism.suggesters.tasks import (
    detect_task_suggestions,
    correct_task_suggestions_in_loop_body,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    create_task_tree,
    __forks,
    set_global_llvm_cxxfilt_path,
    detect_mw_types,
    get_var_definition_line_dict,
)

__global_llvm_cxxfilt_path: str = ""


def build_preprocessed_graph_and_run_detection(
    cu_xml: str,
    dep_file: str,
    reduction_file: str,
    file_mapping: str,
    cu_inst_result_file: str,
    llvm_cxxfilt_path: Optional[str],
    discopop_build_path: Optional[str],
) -> List[PatternInfo]:
    """execute preprocessing of given cu xml file and construct a new cu graph.
    execute run_detection on newly constructed graph afterwards.
    :param cu_xml: Path (string) to the CU xml file to be used
    :param dep_file: Path (string) to the dependence file to be used
    :param loop_counter_file: Path (string) to the loop counter file file to be used
    :param reduction_file: Path (string) to the reduction file to be used
    :param file_mapping: Path (string) to the FileMapping.txt to be used
    :param cu_inst_result_file: Path (string) to the _CUInstResult.txt to be used
    :param llvm_cxxfilt_path: Path (string) to the llvm-cxxfilt executable to be used or None.
    :param discopop_build_path: path (string) to discopop build folder.
    :return: List of detected pattern info
    """
    global __global_llvm_cxxfilt_path
    if llvm_cxxfilt_path is None:
        __global_llvm_cxxfilt_path = "None"
    else:
        __global_llvm_cxxfilt_path = llvm_cxxfilt_path
    if discopop_build_path is None or discopop_build_path == "None":
        raise ValueError("Path to DiscoPoP build directory not specified!")
    set_global_llvm_cxxfilt_path(__global_llvm_cxxfilt_path)
    preprocessed_cu_xml = cu_xml_preprocessing(cu_xml)
    preprocessed_graph = PEGraphX.from_parsed_input(
        *parse_inputs(preprocessed_cu_xml, dep_file, reduction_file, file_mapping)
    )

    # execute reduction detector to enable taskloop-reduction-detection
    detect_reduction(preprocessed_graph)
    detect_do_all(preprocessed_graph)

    suggestions = run_detection(
        preprocessed_graph,
        preprocessed_cu_xml,
        file_mapping,
        dep_file,
        cu_inst_result_file,
        discopop_build_path,
    )

    return suggestions


def run_detection(
    pet: PEGraphX,
    cu_xml: str,
    file_mapping: str,
    dep_file: str,
    cu_ist_result_file: str,
    discopop_build_path: str,
) -> List[PatternInfo]:
    """Computes the Task Parallelism Pattern for a node:
    (Automatic Parallel Pattern Detection in the Algorithm Structure Design Space p.46)
    1.) first merge all children of the node -> all children nodes get the dependencies
        of their children nodes and the list of the children nodes (saved in node.childrenNodes)
    2.) To detect Task Parallelism, we use Breadth First Search (BFS)
        a.) the hotspot becomes a fork
        b.) all child nodes become first worker if they are not marked as worker before
        c.) if a child has dependence to more than one parent node, it will be marked as barrier
    3.) if two barriers can run in parallel they are marked as barrierWorkers.
        Two barriers can run in parallel if there is not a directed path from one to the other
        :param pet: PET graph
        :param cu_xml: Path (string) to the CU xml file to be used
        :param file_mapping: Path (string) to the FileMapping.txt to be used
        :param dep_file: Path (string) to the dependencies-file to be used
        :param cu_ist_result_file: Path(string) to the CUInstResult.txt
        :param discopop_build_path: path to discopop build folder
        :return: List of detected pattern info
    """
    result: List[PatternInfo] = []

    for node in pet.all_nodes():
        if isinstance(node, DummyNode):
            continue
        if pet.direct_children_or_called_nodes(node):
            detect_mw_types(pet, node)

        if node.mw_type == MWType.NONE:
            node.mw_type = MWType.ROOT

    __forks.clear()
    create_task_tree(pet, pet.main)

    fs = [f for f in __forks if f.node_id == "130:0"]

    for fork in fs:
        if fork.child_tasks:
            result.append(TaskParallelismInfo(fork.nodes[0], TPIType.DUMMY, ["dummy_fork"], [], [], [], []))
    # Preprocessing
    check_loop_scopes(pet)
    # Suggestion generation
    result += detect_task_suggestions(pet)
    result += suggest_parallel_regions(pet, cast(List[TaskParallelismInfo], result))
    result = cast(List[PatternInfo], set_task_contained_lines(cast(List[TaskParallelismInfo], result)))
    result = cast(List[PatternInfo], detect_taskloop_reduction(pet, cast(List[TaskParallelismInfo], result)))
    result = cast(
        List[PatternInfo],
        remove_useless_barrier_suggestions(pet, cast(List[TaskParallelismInfo], result)),
    )
    result = detect_barrier_suggestions(pet, result)
    result = validate_barriers(pet, result)
    result = detect_dependency_clauses_alias_based(
        pet, result, file_mapping, cu_xml, dep_file, cu_ist_result_file, discopop_build_path
    )
    result = suggest_missing_barriers_for_global_vars(pet, result)
    result = combine_omittable_cus(pet, result)
    result = suggest_barriers_for_uncovered_tasks_before_return(pet, result)
    result = suggest_shared_clauses_for_all_tasks_in_function_body(pet, result)
    result = remove_duplicates(result)
    result = correct_task_suggestions_in_loop_body(pet, result)
    result = filter_data_sharing_clauses(pet, result, get_var_definition_line_dict(cu_xml))
    result = filter_data_depend_clauses(pet, result, get_var_definition_line_dict(cu_xml))
    result = remove_duplicate_data_sharing_clauses(result)
    result = group_task_suggestions(pet, result)
    result = sort_output(result)

    return result



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast, Dict, Optional, Tuple

from discopop_explorer.PEGraphX import Node, EdgeType, PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    ParallelRegionInfo,
    TPIType,
)


def group_task_suggestions(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """Group task and taskwait suggestions by traversing the successor graph and marking.
    Starting from each taskwait suggestion, the successor graph is traversed in reverse order.
    Visited task suggestions are marked using it's id. The traversal stops if either another taskwait suggestion or the
    end of the graph has been reached.
    After marking, task suggestions which are marked with more than one id denote the presence of a non-trivial
    task group. In a postprocessing step, such tasks are used to combine multiple task groups into a single group.
    The intuition of the result is simple: if any task suggestion within a task group is implemented by the user,
    it is recommended to implement all related suggestions, as they belong together and,
    e.g. in case of missing taskwaits, might influence the correctness of the resulting program.
    :param pet: PET Graph
    :param suggestions: Found suggestions
    :return: Updated suggestions"""
    task_suggestions = [s for s in [e for e in suggestions if type(e) == TaskParallelismInfo] if s.type is TPIType.TASK]
    taskwait_suggestions = [
        s for s in [e for e in suggestions if type(e) == TaskParallelismInfo] if s.type is TPIType.TASKWAIT
    ]
    # mark preceeding suggestions for each taskwait suggestion
    for task_group_id, tws in enumerate(taskwait_suggestions):
        # mark taskwait suggestion with own id
        tws.task_group.append(task_group_id)
        relatives: List[Node] = [tws._node]
        queue: List[Node] = [pet.node_at(in_e[0]) for in_e in pet.in_edges(tws._node.id, EdgeType.SUCCESSOR)]
        while len(queue) > 0:
            cur = queue.pop(0)
            if cur.tp_contains_taskwait:
                continue
            relatives.append(cur)
            for in_edge in pet.in_edges(cur.id, EdgeType.SUCCESSOR):
                if pet.node_at(in_edge[0]) not in relatives + queue:
                    queue.append(pet.node_at(in_edge[0]))
            for out_edge in pet.out_edges(cur.id, EdgeType.SUCCESSOR):
                if pet.node_at(out_edge[1]) not in relatives + queue:
                    queue.append(pet.node_at(out_edge[1]))

        # mark intersection of relatives and task_suggestions
        for intersect in [sug for sug in task_suggestions if sug._node in relatives]:
            intersect.task_group.append(task_group_id)
    # combine groups by replacing taskgroup ids (higher replaced by lower id)
    # get replacements to be done
    replacements: Dict[int, int] = dict()  # [target_id :  replacement_id)]
    for ts in task_suggestions:
        if len(ts.task_group) <= 1:
            continue
        # multiple task_group entries. get smallest id as replacement, add others as targets
        replacement_id = min(ts.task_group)
        for target_id in [tid for tid in ts.task_group]:
            # will lead to identity replacements, required for later replacement
            replacements[target_id] = replacement_id
    # refine replacement list (check for and simplify transitivities)
    modification_found = True
    while modification_found:
        modification_found = False
        for target_id in replacements:
            replacement_id = replacements[target_id]
            if replacement_id == target_id:
                # necessary for completeness of later replacement
                continue
            if replacement_id in replacements:
                # transitive replacement found, simplify
                if replacements[target_id] != replacements[replacement_id]:
                    replacements[target_id] = replacements[replacement_id]
                    modification_found = True
                    break
    # execute replacement
    for sug in task_suggestions + taskwait_suggestions:
        sug.task_group = [replacements[tg_elem] if tg_elem in replacements else tg_elem for tg_elem in sug.task_group]
        # validate and combine results
        # valid, if all entries in sug.task_group are equal. Replace by single entry if valid.
        value: Optional[int] = None
        for tg_elem in sug.task_group:
            if value is None:
                value = tg_elem
                continue
            if tg_elem != value:
                # not valid, raise exception
                raise ValueError(
                    "Task Group creation led to erroneous results: NodeId:",
                    sug._node.id,
                    "  Task Group: ",
                    sug.task_group,
                )
        # valid, overwrite sug.task_group if value is not None
        if value is not None:
            sug.task_group = [value]
    return suggestions


def sort_output(suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """orders the list of suggestions by the respective properties:
    order by: file-id, then line-number (descending).
    Returns the sorted list of suggestions
    :param suggestions: List[PatternInfo]
    :return: List[PatternInfo]
    """
    sorted_suggestions = []
    tmp_dict: Dict[str, List[Tuple[str, PatternInfo]]] = dict()
    for sug in suggestions:
        if isinstance(sug, (ParallelRegionInfo, TaskParallelismInfo)):
            # get start_line and file_id for sug
            if ":" not in sug.region_start_line:
                start_line = sug.region_start_line
                file_id = sug.start_line[0 : sug.start_line.index(":")]
            else:
                start_line = sug.region_start_line
                file_id = start_line[0 : start_line.index(":")]
                start_line = start_line[start_line.index(":") + 1 :]
            # split suggestions by file-id
            if file_id not in tmp_dict:
                tmp_dict[file_id] = []
            tmp_dict[file_id].append((start_line, sug))
        else:
            continue
    # sort suggestions by line-number (descending)
    for key in tmp_dict:
        sorted_list = cast(
            List[Tuple[str, PatternInfo]],
            sorted(tmp_dict[key], key=lambda x: int(x[0]), reverse=True),
        )
        sorted_list_pruned = cast(List[PatternInfo], [elem[1] for elem in sorted_list])
        sorted_suggestions += sorted_list_pruned
    return sorted_suggestions



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
import shutil
import subprocess
from pathlib import Path
from typing import Dict, List

from discopop_library.LineMapping.diff_modifications import apply_line_mapping_modifications_from_files
from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments


def apply_patches(
    apply: List[str],
    file_mapping: Dict[int, Path],
    arguments: PatchApplicatorArguments,
    applied_suggestions_file: str,
    patch_generator_dir: str,
) -> int:
    """Return values:"
    "0: Applied successfully"
    "1: Nothing applied"
    "2: Some changes applied successfully"""
    retval = -1  # -1 -> nothing seen so far
    # get list of applicable suggestions
    applicable_suggestions = [name for name in os.listdir(patch_generator_dir)]

    # get already applied suggestions
    with open(applied_suggestions_file, "r") as f:
        applied_suggestions = json.loads(f.read())
        if arguments.verbose:
            print("Previously applied suggestions: ", applied_suggestions["applied"])

    for suggestion_id in apply:
        if suggestion_id in applied_suggestions["applied"]:
            if arguments.verbose:
                print("Skipping already applied suggestion: ", suggestion_id)
            continue
        if suggestion_id in applicable_suggestions:
            if arguments.verbose:
                print("Applying suggestion ", suggestion_id)
            successful = __apply_file_patches(file_mapping, suggestion_id, patch_generator_dir, arguments)
            if successful:
                applied_suggestions["applied"].append(suggestion_id)
                # write updated applied suggestions to file
                with open(applied_suggestions_file, "w") as f:
                    f.write(json.dumps(applied_suggestions))
                # update return code
                if retval == -1:
                    retval = 0
                if retval == 1:
                    # no update for retval = 0 necessary
                    retval = 2
            else:
                print("Applying suggestion", suggestion_id, "not successful.")
                # update return code
                if retval == -1:
                    retval = 1
                if retval == 0:
                    retval = 2
        else:
            if arguments.verbose:
                print("Nothing to apply for suggestion ", suggestion_id)
    if retval == -1:
        retval = 0
    return retval


def __apply_file_patches(
    file_mapping: Dict[int, Path], suggestion_id: str, patch_generator_dir: str, arguments: PatchApplicatorArguments
) -> bool:
    # get a list of patches for the given suggestion
    patch_files = os.listdir(os.path.join(patch_generator_dir, suggestion_id))
    if arguments.verbose:
        print("\tFound patch files:", patch_files)
    encountered_error = False
    already_patched: List[str] = []
    for patch_file_name in patch_files:
        patch_file_id = int(patch_file_name.rstrip(".patch"))
        patch_target = file_mapping[patch_file_id]
        patch_file_path = os.path.join(patch_generator_dir, suggestion_id, patch_file_name)
        # save original version to calculate diff to calculate line mapping
        shutil.copyfile(patch_target.as_posix(), patch_target.as_posix() + ".line_mapping_tmp")
        command = [
            "patch",
            patch_target.as_posix(),
            patch_file_path,
        ]
        if arguments.verbose:
            print("\tapplying: ", " ".join(command))
        result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=os.getcwd(),
        )
        if result.returncode != 0:
            if arguments.verbose:
                print("RESULT: ", result.returncode)
                print("STDERR:")
                print(result.stderr)
            print("STDOUT: ")
            print(result.stdout)
            encountered_error = True
            # delete temporary file
            os.remove(patch_target.as_posix() + ".line_mapping_tmp")
            break
        else:
            already_patched.append(patch_file_name)
            # apply modifications to line mapping
            apply_line_mapping_modifications_from_files(
                patch_file_id, patch_target.as_posix() + ".line_mapping_tmp", patch_target.as_posix()
            )
            # delete temporary file
            os.remove(patch_target.as_posix() + ".line_mapping_tmp")

    # cleanup in case of an error
    if encountered_error:
        for patch_file_name in already_patched:
            patch_file_id = int(patch_file_name.rstrip(".patch"))
            patch_target = file_mapping[patch_file_id]
            patch_file_path = os.path.join(patch_generator_dir, suggestion_id, patch_file_name)
            # save original version to calculate diff to calculate line mapping
            shutil.copyfile(patch_target.as_posix(), patch_target.as_posix() + ".line_mapping_tmp")
            command = [
                "patch",
                "-R",
                patch_target.as_posix(),
                patch_file_path,
            ]
            if arguments.verbose:
                print("\tcleanup: applying: ", " ".join(command))
            result = subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                cwd=os.getcwd(),
            )
            if result.returncode != 0:
                if arguments.verbose:
                    print("RESULT: ", result.returncode)
                    print("STDERR:")
                    print(result.stderr)
                print("STDOUT: ")
                print(result.stdout)

            # apply modifications to line mapping
            apply_line_mapping_modifications_from_files(
                patch_file_id, patch_target.as_posix() + ".line_mapping_tmp", patch_target.as_posix()
            )
            # delete temporary file
            os.remove(patch_target.as_posix() + ".line_mapping_tmp")

    return not encountered_error



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
import tkinter as tk
from tkinter import ttk
from typing import List

from DEPRECATED_discopop_wizard.classes.Suggestion import Suggestion
from DEPRECATED_discopop_wizard.screens.utils import create_tool_tip
from DEPRECATED_discopop_wizard.screens.widgets.ScrollableText import ScrollableTextWidget
from DEPRECATED_discopop_wizard.utils import support_scrolling


def show_suggestions_overview_screen(wizard, details_frame: tk.Frame, execution_configuration_obj):
    # close elements on details_frame
    for c in details_frame.winfo_children():
        c.destroy()
    # load suggestions from execution
    suggestions = get_suggestion_objects(wizard, execution_configuration_obj)

    # create horizontally split frames (scrollable list of suggestions + code preview)
    horizontal_paned_window = ttk.PanedWindow(details_frame, orient="horizontal")
    horizontal_paned_window.pack(fill=tk.BOTH, expand=True)
    scrollable_list_frame = tk.Frame(horizontal_paned_window)
    horizontal_paned_window.add(scrollable_list_frame, weight=1)
    code_preview_frame = tk.Frame(horizontal_paned_window)
    horizontal_paned_window.add(code_preview_frame, weight=5)

    tmp_frame = tk.Frame(scrollable_list_frame)
    tmp_frame.pack(fill=tk.BOTH, expand=True)

    # define notebook widget to show generated CUs and dependencies
    separator = ttk.Separator(scrollable_list_frame, orient="horizontal")
    separator.pack(fill=tk.X)
    result_browser_frame = tk.Frame(scrollable_list_frame)
    result_browser_frame.pack(fill=tk.BOTH, expand=True)
    result_notebook = ttk.Notebook(result_browser_frame)
    result_notebook.pack(fill=tk.BOTH, expand=True)
    # add CU preview
    cu_display_widget = ScrollableTextWidget(result_notebook)
    with open(execution_configuration_obj.value_dict["working_copy_path"] + "/Data.xml", "r") as f:
        cu_display_widget.set_text(f.read())
    result_notebook.add(cu_display_widget.frame, text="CU's")
    # add dynamic dependency preview
    dynamic_dep_display_widget = ScrollableTextWidget(result_notebook)
    with open(
        execution_configuration_obj.value_dict["working_copy_path"]
        + "/"
        + execution_configuration_obj.value_dict["executable_name"]
        + "_dp_dep.txt",
        "r",
    ) as f:
        dynamic_dep_display_widget.set_text(f.read())
    result_notebook.add(dynamic_dep_display_widget.frame, text="Dynamic DEP's")
    # add static dependency preview
    static_dep_display_widget = ScrollableTextWidget(result_notebook)
    if os.path.exists(execution_configuration_obj.value_dict["working_copy_path"] + "/static_dependencies.txt"):
        with open(
            execution_configuration_obj.value_dict["working_copy_path"] + "/static_dependencies.txt",
            "r",
        ) as f:
            static_dep_display_widget.set_text(f.read())
    else:
        static_dep_display_widget.set_text(
            execution_configuration_obj.value_dict["working_copy_path"] + "/static_dependencies.txt" + " NOT FOUND."
        )
    result_notebook.add(static_dep_display_widget.frame, text="Static DEP's")
    # add instrumented LLVM IR preview
    instrumented_llvm_ir_display_widget = ScrollableTextWidget(result_notebook)
    if os.path.exists(
        execution_configuration_obj.value_dict["working_copy_path"]
        + "/"
        + execution_configuration_obj.value_dict["executable_name"]
        + "_dp.ll"
    ):
        with open(
            execution_configuration_obj.value_dict["working_copy_path"]
            + "/"
            + execution_configuration_obj.value_dict["executable_name"]
            + "_dp.ll",
            "r",
        ) as f:
            instrumented_llvm_ir_display_widget.set_text(f.read())
    else:
        instrumented_llvm_ir_display_widget.set_text(
            execution_configuration_obj.value_dict["working_copy_path"]
            + "/"
            + execution_configuration_obj.value_dict["executable_name"]
            + "_dp.ll"
            + " NOT FOUND."
        )
    result_notebook.add(instrumented_llvm_ir_display_widget.frame, text="LLVM IR")
    # add patterns.json preview
    patterns_json_display_widget = ScrollableTextWidget(result_notebook)
    with open(execution_configuration_obj.value_dict["working_copy_path"] + "/" + "patterns.json", "r") as f:
        patterns_json_display_widget.set_text(f.read())
    result_notebook.add(patterns_json_display_widget.frame, text="patterns.json")

    # create scrollable list of suggestions
    canvas = tk.Canvas(tmp_frame)
    scrollbar = tk.Scrollbar(tmp_frame, orient="vertical", command=canvas.yview)
    scrollable_frame = tk.Frame(canvas)
    scrollable_frame.bind("<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all")))
    canvas.create_window((0, 0), window=scrollable_frame, anchor="nw")
    canvas.configure(yscrollcommand=scrollbar.set)
    for row, suggestion in enumerate(suggestions):
        # create button to load code preview
        button = suggestion.get_as_button(scrollable_frame, code_preview_frame, execution_configuration_obj)
        button.pack(fill=tk.BOTH, expand=True)

        # register hover message (suggestion details)
        create_tool_tip(button, text=suggestion.get_details())

    # add support for mouse wheel scrolling
    support_scrolling(canvas)

    # add label
    tk.Label(tmp_frame, text="Suggestions", font=wizard.style_font_bold).pack(side="top", pady=10)

    canvas.pack(side="left", fill="both", expand=True)
    scrollbar.pack(side="right", fill="y")


def get_suggestion_objects(wizard, execution_configuration_obj) -> List[Suggestion]:
    suggestions_path = os.path.join(execution_configuration_obj.value_dict["working_copy_path"], "patterns.json")

    suggestions_list: List[Suggestion] = []
    with open(suggestions_path, "r") as f:
        suggestions_dict = json.load(f)
    for suggestion_type in suggestions_dict:
        for suggestion_values in suggestions_dict[suggestion_type]:
            suggestions_list.append(Suggestion(wizard, suggestion_type, suggestion_values))

    return suggestions_list



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast

from discopop_explorer.PEGraphX import PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.combined_gpu_patterns.CombinedGPURegions import (
    find_combined_gpu_regions,
)


def run_detection(pet: PEGraphX, res, project_folder_path: str) -> List[PatternInfo]:
    """Search for combined gpu patterns

    :param pet: PET graph
    :param res: DetectionResult object
    :return: List of detected pattern info
    """

    # construct Combined GPU Regions
    combined_gpu_regions = find_combined_gpu_regions(pet, res.simple_gpu, project_folder_path)

    return cast(List[PatternInfo], combined_gpu_regions)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import sys
from typing import List, Tuple, Dict, Set

from discopop_explorer.PEGraphX import EdgeType, CUNode, PEGraphX, NodeID, MemoryRegion
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Dependency import Dependency
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import (
    ExitPointPositioning,
    EntryPointPositioning,
    ExitPointType,
    EntryPointType,
    UpdateType,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.prepare_metadata import (
    get_dependencies_as_metadata,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_1 import (
    get_written_and_read_memory_regions_by_cu,
    get_cu_and_varname_to_memory_regions,
    get_memory_region_to_cu_and_variables_dict,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_2 import (
    populate_live_data,
    add_memory_regions_to_device_liveness,
    propagate_memory_regions,
    convert_liveness,
    calculate_host_liveness,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_3 import (
    initialize_writes,
    cleanup_writes,
    group_writes_by_cu,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_4 import (
    create_circle_free_function_graphs,
    add_accesses_from_called_functions,
    identify_updates_in_unrolled_function_graphs,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_5 import (
    propagate_variable_name_associations,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.step_6 import (
    convert_updates_to_entry_and_exit_points,
    identify_end_of_life_points,
    add_aliases,
    extend_region_liveness_using_unrolled_functions,
    remove_duplicates,
    join_elements,
)
from discopop_explorer.pattern_detectors.simple_gpu_patterns.GPURegions import GPURegionInfo


class CombinedGPURegion(PatternInfo):
    contained_regions: List[GPURegionInfo]
    update_instructions: List[
        Tuple[NodeID, NodeID, UpdateType, str, str]
    ]  # (source_cu_id, sink_cu_id, UpdateType, target_vars, meta_line_num)
    target_data_regions: Dict[str, List[Tuple[List[NodeID], NodeID, NodeID, str, str]]]
    # {var: ([contained cu_s], entry_cu, exit_after_cu, meta_entry_line_num, meta_exit_line_num)
    data_region_entry_points: List[
        Tuple[VarName, NodeID, EntryPointType, str, EntryPointPositioning]
    ]  # [(var, cu_id, entry_point_type, meta_line_num, positioning)]
    data_region_exit_points: List[
        Tuple[VarName, NodeID, ExitPointType, str, ExitPointPositioning]
    ]  # [(var, cu_id, exit_point_type, meta_line_num, positioning)]
    data_region_depend_in: List[Tuple[VarName, NodeID, str]]  # [(var, cu_id, meta_line_num)]
    data_region_depend_out: List[Tuple[VarName, NodeID, str]]  # [(var, cu_id, meta_line_num)]
    device_cu_ids: List[NodeID]
    # meta information, mainly for display and overview purposes
    meta_device_lines: List[str]
    meta_host_lines: List[str]
    meta_device_liveness: Dict[MemoryRegion, List[str]]
    meta_host_liveness: Dict[MemoryRegion, List[str]]
    project_folder_path: str

    def __init__(self, pet: PEGraphX, contained_regions: List[GPURegionInfo], project_folder_path: str):
        self.project_folder_path = project_folder_path
        node_id = sorted([region.node_id for region in contained_regions])[0]
        device_cu_ids: List[NodeID] = []
        for region in contained_regions:
            device_cu_ids += [NodeID(cu_id_str) for cu_id_str in region.contained_cu_ids]
            device_cu_ids = list(set(device_cu_ids))
        PatternInfo.__init__(self, pet.node_at(node_id))
        self.contained_regions = contained_regions
        self.device_cu_ids = device_cu_ids
        self.start_line = min(
            [l.start_line for l in contained_regions]
        )  # todo: not correct anymore since multiple files contained now
        self.end_line = max(
            [l.end_line for l in contained_regions]
        )  # todo: not correct anymore since multiple files contained now
        print("\n\n", file=sys.stderr)
        print("DEVICE CU IDS: ", file=sys.stderr)
        print(self.device_cu_ids, file=sys.stderr)

        # initialize object
        self.update_instructions = []
        self.data_region_entry_points = []
        self.data_region_exit_points = []
        self.data_region_depend_in = []
        self.data_region_depend_out = []
        self.meta_device_liveness: Dict[MemoryRegion, List[str]] = dict()
        self.meta_host_liveness: Dict[MemoryRegion, List[str]] = dict()
        self.meta_host_lines: Dict[VarName, List[str]]

        # ### STEP 1: INITIALIZATION
        # get written and read memory regions by CU
        written_memory_regions_by_cu: Dict[NodeID, Set[MemoryRegion]]
        read_memory_regions_by_cu: Dict[NodeID, Set[MemoryRegion]]
        (
            written_memory_regions_by_cu,
            read_memory_regions_by_cu,
        ) = get_written_and_read_memory_regions_by_cu(self.contained_regions, pet)

        # get memory region and variable associations for each CU
        cu_and_variable_to_memory_regions: Dict[
            NodeID, Dict[VarName, Set[MemoryRegion]]
        ] = get_cu_and_varname_to_memory_regions(self.contained_regions, pet, written_memory_regions_by_cu)

        print("WRITTEN MEMORY REGIONS BY CU: ", file=sys.stderr)
        print(written_memory_regions_by_cu, file=sys.stderr)
        print(file=sys.stderr)

        print("READ MEMORY REGIONS BY CU: ", file=sys.stderr)
        print(read_memory_regions_by_cu, file=sys.stderr)
        print(file=sys.stderr)

        # get memory regions to cus and variables names
        memory_regions_to_cus_and_variables: Dict[
            MemoryRegion, Dict[NodeID, Set[VarName]]
        ] = get_memory_region_to_cu_and_variables_dict(cu_and_variable_to_memory_regions)
        print("MEMORY REGIONS TO CUS AND VARIABLES:", file=sys.stderr)
        print(memory_regions_to_cus_and_variables, file=sys.stderr)
        print(file=sys.stderr)

        # ### STEP 2.1: INITIALIZE LIVE DATA USING MAPPING CLAUSES AND CONVERSION TO MEMORY REGIONS
        live_device_variables = populate_live_data(self, pet, ignore_update_instructions=True)
        print("LIVE DEVICE VARIABLES:", file=sys.stderr)
        print(live_device_variables, file=sys.stderr)
        print(file=sys.stderr)

        # extend device liveness with memory regions
        device_liveness_plus_memory_regions: Dict[
            VarName, List[Tuple[NodeID, Set[MemoryRegion]]]
        ] = add_memory_regions_to_device_liveness(live_device_variables, cu_and_variable_to_memory_regions)

        # ### STEP 2.2: CALCULATE LIVE DATA BY PROPAGATING MEMORY REGIONS AND EXTENDING LIFESPAN

        # propagate memory regions
        device_liveness_plus_memory_regions = propagate_memory_regions(device_liveness_plus_memory_regions)

        print("PROPAGATED DEVICE VARIABLES + MEMORY:", file=sys.stderr)
        print(device_liveness_plus_memory_regions, file=sys.stderr)
        print(file=sys.stderr)

        # convert liveness to memory regions as basis
        memory_region_liveness = convert_liveness(device_liveness_plus_memory_regions)
        print("CONVERTED DEVICE MEMORY REGION LIVENESS:", file=sys.stderr)
        print(memory_region_liveness, file=sys.stderr)
        print(file=sys.stderr)

        # extend data liveness
        extended_memory_region_liveness = memory_region_liveness  # extend_data_lifespan(pet, memory_region_liveness)

        print("EXTENDED DEVICE MEMORY REGION LIVENESS:", file=sys.stderr)
        print(extended_memory_region_liveness, file=sys.stderr)
        print(file=sys.stderr)

        # ### STEP 2.3: CALCULATE HOST LIVENESS
        host_liveness = calculate_host_liveness(self, pet)
        print("HOST LIVENESS:", file=sys.stderr)
        print(host_liveness, file=sys.stderr)
        print(file=sys.stderr)
        host_memory_region_liveness = convert_liveness(host_liveness)
        extended_host_memory_region_liveness = (
            host_memory_region_liveness  # extend_data_lifespan(           pet, host_memory_region_liveness        )
        )
        print("EXTENDED HOST MEMORY REGION LIVENESS:", file=sys.stderr)
        print(extended_host_memory_region_liveness, file=sys.stderr)
        print(file=sys.stderr)

        # ### STEP 3: MARK WRITTEN VARIABLES
        # initialize writes
        device_writes = initialize_writes(extended_memory_region_liveness, written_memory_regions_by_cu)
        host_writes = initialize_writes(extended_host_memory_region_liveness, written_memory_regions_by_cu)

        # propagate writes to parents, successors and the children of successors
        propagated_device_writes = device_writes  # propagate_writes(self, pet, device_writes)
        print("PROPAGATED DEVICE WRITES:", file=sys.stderr)
        print(propagated_device_writes, file=sys.stderr)
        print(file=sys.stderr)

        propagated_host_writes = host_writes  # propagate_writes(self, pet, host_writes)
        print("PROPAGATED HOST WRITES:", file=sys.stderr)
        print(propagated_host_writes, file=sys.stderr)
        print(file=sys.stderr)

        # cleanup propagated writes (remove None entries if they are overwritten by at least one write)
        propagated_device_writes = cleanup_writes(propagated_device_writes)
        propagated_host_writes = cleanup_writes(propagated_host_writes)
        print("CLEANED HOST WRITES:", file=sys.stderr)
        print(propagated_host_writes, file=sys.stderr)
        print(file=sys.stderr)

        # group by cus
        device_writes_by_cu = group_writes_by_cu(propagated_device_writes)
        print("DEVICE WRITES BY CU:", file=sys.stderr)
        print(device_writes_by_cu, file=sys.stderr)
        print(file=sys.stderr)

        host_writes_by_cu = group_writes_by_cu(propagated_host_writes)
        print("HOST WRITES BY CU:", file=sys.stderr)
        print(host_writes_by_cu, file=sys.stderr)
        print(file=sys.stderr)

        # ### STEP 4: IDENTIFY SYNCHRONOUS UPDATE POINTS
        writes_by_device = {0: host_writes_by_cu, 1: device_writes_by_cu}

        # unroll function bodies to create circle-free graphs
        unrolled_function_graphs = create_circle_free_function_graphs(pet, add_dummy_node=False)
        # TODO add accesses from called function to the calling CUs
        writes_by_device = add_accesses_from_called_functions(
            pet, writes_by_device, force_called_functions_to_host=True
        )

        # identify updates based on the circle-free graph
        # TODO parallelize on function level

        # issued_updates = identify_updates(self, pet, writes_by_device)  # old code
        issued_updates = identify_updates_in_unrolled_function_graphs(
            self, pet, writes_by_device, unrolled_function_graphs
        )

        print("ISSUED UPDATES:", file=sys.stderr)
        for update in issued_updates:
            print(update, file=sys.stderr)
        print(file=sys.stderr)

        # remove dummy marks from CU ID's created during loop unrolling
        for update in issued_updates:
            update.remove_dummy_marks()

        # ### STEP 5: CONVERT MEMORY REGIONS IN UPDATES TO VARIABLE NAMES
        # propagate memory region to variable name associations within function body
        memory_regions_to_functions_and_variables: Dict[
            MemoryRegion, Dict[NodeID, Set[VarName]]
        ] = propagate_variable_name_associations(pet, memory_regions_to_cus_and_variables)
        print("MEMORY REGIONS TO FUNCTIONS AND VARIABLES:", file=sys.stderr)
        print(memory_regions_to_functions_and_variables, file=sys.stderr)
        print(file=sys.stderr)

        # determine variable name for memory regions in update instructions
        for update in issued_updates:
            update.convert_memory_regions_to_variable_names(pet, memory_regions_to_functions_and_variables)

        # ### POTENTIAL STEP 6: CONVERT MEMORY REGIONS TO STRUCTURE INDICES

        # ### STEP 6: convert updates to entry / exit points if possible

        # extend memory region liveness based on unrolled function graphs
        extended_memory_region_liveness = extend_region_liveness_using_unrolled_functions(
            pet, extended_memory_region_liveness, unrolled_function_graphs
        )
        extended_host_memory_region_liveness = extend_region_liveness_using_unrolled_functions(
            pet, extended_host_memory_region_liveness, unrolled_function_graphs
        )

        # add aliases to updates
        aliased_updates = add_aliases(
            pet,
            issued_updates,
            memory_regions_to_functions_and_variables,
        )

        memory_region_liveness_by_device = {
            0: extended_host_memory_region_liveness,
            1: extended_memory_region_liveness,
        }
        entry_points, exit_points, updates = convert_updates_to_entry_and_exit_points(
            pet, aliased_updates, memory_region_liveness_by_device
        )

        # identify exit points
        exit_points.update(
            identify_end_of_life_points(
                pet,
                entry_points,
                exit_points,
                memory_region_liveness_by_device,
                memory_regions_to_functions_and_variables,
            )
        )

        # remove duplicates
        updates = remove_duplicates(updates)
        entry_points = remove_duplicates(entry_points)
        exit_points = remove_duplicates(exit_points)

        # join entries
        updates = join_elements(updates)
        entry_points = join_elements(entry_points)
        exit_points = join_elements(exit_points)

        # ### PREPARE METADATA
        # prepare device liveness
        #        self.meta_device_liveness = prepare_liveness_metadata(
        #            pet,
        #            extended_memory_region_liveness,
        #            propagated_device_writes,
        #            self.meta_device_liveness,
        #        )
        # prepare host liveness
        #        self.meta_host_liveness = prepare_liveness_metadata(
        #            pet,
        #            extended_host_memory_region_liveness,
        #            propagated_host_writes,
        #            self.meta_host_liveness,
        #        )
        # prepare update instructions
        self.update_instructions = [
            update.get_as_metadata_using_variable_names(pet, self.project_folder_path) for update in updates
        ]
        # prepare entry points
        self.data_region_entry_points = [
            entry_point.get_as_metadata(pet, self.project_folder_path) for entry_point in entry_points
        ]

        # prepare exit points
        self.data_region_exit_points = [
            exit_point.get_as_metadata(pet, self.project_folder_path) for exit_point in exit_points
        ]

        # prepare dependencies
        all_dependencies: Set[Dependency] = set()
        for update in updates:
            all_dependencies.update(update.dependencies)
        for entry_point in entry_points:
            all_dependencies.update(entry_point.dependencies)
        for exit_point in exit_points:
            all_dependencies.update(exit_point.dependencies)

        self.data_region_depend_in, self.data_region_depend_out = get_dependencies_as_metadata(pet, all_dependencies)

    def __str__(self):
        raise NotImplementedError()  # used to identify necessity to call to_string() instead

    def to_string(self, pet: PEGraphX):
        contained_regions_str = "\n" if len(self.contained_regions) > 0 else ""
        for region in self.contained_regions:
            region_str = region.to_string(pet)
            # pretty printing
            region_str = "".join(["\t" + s + "\n" for s in region_str.split("\n")])
            contained_regions_str += region_str

        return (
            f"COMBINED GPU Region at: {self.node_id}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"Device CUs: {self.device_cu_ids}\n"
            f"Device lines: {self.meta_device_lines}\n"
            f"Host lines: {self.meta_host_lines}\n"
            f"Contained regions: {contained_regions_str}\n"
        )


def find_combined_gpu_regions(
    pet: PEGraphX, gpu_regions: List[GPURegionInfo], project_folder_path: str
) -> List[CombinedGPURegion]:
    # create combined gpu regions from original gpu regions
    combined_gpu_regions = []
    for gpu_region in gpu_regions:
        combined_gpu_regions.append(CombinedGPURegion(pet, [gpu_region], project_folder_path))

    # determine relations between single-element regions
    combinable_pairs: List[Tuple[CombinedGPURegion, CombinedGPURegion]] = find_all_pairwise_gpu_region_combinations(
        combined_gpu_regions, pet
    )

    intra_function_combinations = find_combinations_within_function_body(pet, combinable_pairs)

    true_successor_combinations = find_true_successor_combinations(pet, intra_function_combinations)

    # combine regions
    for combinable_1, combinable_2 in intra_function_combinations:  # true_successor_combinations:
        if combinable_1 in combined_gpu_regions:
            combined_gpu_regions.remove(combinable_1)
        if combinable_2 in combined_gpu_regions:
            combined_gpu_regions.remove(combinable_2)
        combined_gpu_regions.append(combine_regions(pet, combinable_1, combinable_2, project_folder_path))

    #    # combine all known regions
    #    combined_region: Optional[CombinedGPURegion] = None
    #    for cgr in combined_gpu_regions:
    #        if combined_region is None:
    #            combined_region = cgr
    #        else:
    #            combined_region = combine_regions(pet, combined_region, cgr, project_folder_path)

    #    if combined_region is None:
    #        return []
    #    else:
    #        return [combined_region]

    return combined_gpu_regions


def find_all_pairwise_gpu_region_combinations(
    gpu_regions: List[CombinedGPURegion], pet: PEGraphX
) -> List[Tuple[CombinedGPURegion, CombinedGPURegion]]:
    combinable_pairs: List[Tuple[CombinedGPURegion, CombinedGPURegion]] = []  # [(region1, region2)
    # get all pairwise combinations of gpu regions
    for gpu_region_1 in gpu_regions:
        for gpu_region_2 in gpu_regions:
            if gpu_region_1 == gpu_region_2:
                continue
            combinable_pairs.append((gpu_region_1, gpu_region_2))
    return combinable_pairs


def find_combinations_within_function_body(
    pet: PEGraphX, combinable_pairs: List[Tuple[CombinedGPURegion, CombinedGPURegion]]
) -> List[Tuple[CombinedGPURegion, CombinedGPURegion]]:
    """Check regions pairwise for reachability via successor edges.
    Only combinations within a function's body are possible in this way since successor edges only exist
    for function body's."""
    result = []
    for region_1, region_2 in combinable_pairs:
        if region_1 == region_2:
            continue
        # check reachability in both directions via successor edges
        # consider children, as loop nodes do not have successors on their own
        for region_1_child in pet.direct_children(pet.node_at(region_1.contained_regions[0].node_id)):
            for region_2_child in pet.direct_children(pet.node_at(region_2.contained_regions[0].node_id)):
                if region_1_child == region_2_child:
                    continue
                if pet.check_reachability(region_2_child, region_1_child, [EdgeType.SUCCESSOR]):
                    result.append((region_1, region_2))
                elif pet.check_reachability(region_1_child, region_2_child, [EdgeType.SUCCESSOR]):
                    result.append((region_2, region_1))
    # remove duplicates (deterministic ordering not relevant, hence list(set(..)) is fine)
    result = list(set(result))
    return result


def find_true_successor_combinations(
    pet, intra_function_combinations: List[Tuple[CombinedGPURegion, CombinedGPURegion]]
) -> List[Tuple[CombinedGPURegion, CombinedGPURegion]]:
    """Check for combinations options without branching inbetween.
    As a result, both regions will always be executed in succession."""
    result = []
    # a successor path between region_1 and region_2 must exist.
    # a true successor relation exists, if every successor path outgoing from any child of region_1 arrives at region_2
    for region_1, region_2 in intra_function_combinations:
        true_successors = True
        queue: List[CUNode] = pet.direct_children(pet.node_at(region_1.contained_regions[0].node_id))
        visited: List[CUNode] = []
        while queue:
            current_node: CUNode = queue.pop()
            visited.append(current_node)
            if current_node in pet.direct_children(pet.node_at(region_2.contained_regions[0].node_id)):
                # reached region_2
                continue
            # region_2 not reached
            successors = pet.direct_successors(current_node)
            if len(successors) == 0:
                # end of the function body reached, region_2 not reached
                true_successors = False
                break
            else:
                # end of the function's body not yet reached, continue searching
                # add successors to queue
                queue += [succ for succ in successors if succ not in visited]
        if true_successors:
            result.append((region_1, region_2))
    return result


def combine_regions(
    pet: PEGraphX,
    region_1: CombinedGPURegion,
    region_2: CombinedGPURegion,
    project_folder_path: str,
) -> CombinedGPURegion:
    """Combines regions. Individual contained regions are not yet merged!
    Analysis of Live-data and necessary update pragmas needs ti happen in a subsequent step."""
    combined_region = CombinedGPURegion(
        pet, region_1.contained_regions + region_2.contained_regions, project_folder_path
    )
    return combined_region



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast, Tuple, Any

from discopop_explorer.PEGraphX import Node, CUNode, EdgeType, NodeType, PEGraphX, LineID
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    ParallelRegionInfo,
    OmittableCuInfo,
    TaskParallelismInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    check_reachability,
    line_contained_in_region,
    get_predecessor_nodes,
)


def detect_barrier_suggestions(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """detect barriers which have not been detected by __detect_mw_types,
    especially marks WORKER as BARRIER_WORKER if it has depencies to two or
    more CUs which are contained in a path to a CU containing at least one
    suggested Task.
    If omittable CUs are found in the process, they will be marked in the
    pet graph and an intermediate entry in suggestions will be created.
    function executed is repeated until convergence.
    steps:
    1.) mark node as Barrier, if dependences only to task-containing-paths
    :param pet: PET Graph
    :param suggestions: List[TaskParallelismInfo]
    :return List[PatternInfo]
    """
    # split suggestions into task and taskwait suggestions
    taskwait_suggestions: List[TaskParallelismInfo] = []
    task_suggestions: List[TaskParallelismInfo] = []
    omittable_suggestions: List[PatternInfo] = []
    __split_suggestions(suggestions, taskwait_suggestions, task_suggestions, omittable_suggestions)

    for s in task_suggestions:
        s._node.tp_contains_task = True
    for s in taskwait_suggestions:
        s._node.tp_contains_taskwait = True
    task_nodes = [t._node for t in task_suggestions]
    barrier_nodes = [t._node for t in taskwait_suggestions]
    omittable_nodes: List[Tuple[Node, List[Node]]] = []

    transformation_happened = True
    # let run until convergence
    queue = list(pet.all_nodes())
    while transformation_happened or len(queue) > 0:
        transformation_happened = False
        v = queue.pop(0)
        # check step 1
        out_dep_edges = [
            (s, t, e) for s, t, e in pet.out_edges(v.id) if e.etype == EdgeType.DATA and pet.node_at(t) != v
        ]
        # ignore cyclic dependencies on the same variable
        to_remove = []
        for dep_edge in out_dep_edges:
            targets_cyclic_dep_edges = [
                (s, t, e)
                for s, t, e in pet.out_edges(dep_edge[1])
                if e.etype == EdgeType.DATA and t == dep_edge[0] and e.var_name == dep_edge[2].var_name
            ]
            if len(targets_cyclic_dep_edges) != 0:
                to_remove.append(dep_edge)
        for e in to_remove:
            out_dep_edges.remove(e)

        v_first_line = v.start_position()
        v_first_line = LineID(v_first_line[v_first_line.index(":") + 1 :])

        task_count, barrier_count, omittable_count, normal_count = __count_adjacent_nodes(
            pet, suggestions, out_dep_edges, task_nodes, barrier_nodes, omittable_nodes
        )
        if task_count == 1 and barrier_count == 0:
            if not v.tp_omittable:
                # actual change
                v.tp_omittable = True
                combine_with_node_list = [pet.node_at(e[1]) for e in out_dep_edges if pet.node_at(e[1]) in task_nodes]
                if len(combine_with_node_list) < 1:
                    raise ValueError("length combine_with_node < 1!")
                combine_with_node = combine_with_node_list[0]
                omittable_nodes.append((v, [combine_with_node]))
                suggestions.append(OmittableCuInfo(v, combine_with_node))
                transformation_happened = True
        elif barrier_count != 0 and task_count != 0:
            # check if child barrier(s) cover each child task
            child_barriers = [e[1] for e in out_dep_edges if pet.node_at(e[1]).tp_contains_taskwait is True]
            child_tasks = [pet.node_at(e[1]) for e in out_dep_edges if pet.node_at(e[1]).tp_contains_task is True]
            uncovered_task_exists = False
            for ct in child_tasks:
                ct_start_line = ct.start_position()
                ct_start_line = LineID(ct_start_line[ct_start_line.index(":") + 1 :])
                ct_end_line = ct.end_position()
                ct_end_line = LineID(ct_end_line[ct_end_line.index(":") + 1 :])
                # check if ct covered by a barrier
                for cb_id in child_barriers:
                    cb = pet.node_at(cb_id)
                    cb_start_line = cb.start_position()
                    cb_start_line = LineID(cb_start_line[cb_start_line.index(":") + 1 :])
                    cb_end_line = cb.end_position()
                    cb_end_line = LineID(cb_end_line[cb_end_line.index(":") + 1 :])
                    if not (cb_start_line > ct_start_line and cb_end_line > ct_end_line):
                        uncovered_task_exists = True
            if uncovered_task_exists:
                # suggest barrier
                if v.tp_contains_taskwait is False:
                    # actual change
                    v.tp_contains_taskwait = True
                    barrier_nodes.append(v)
                    transformation_happened = True
                    tmp_suggestion = TaskParallelismInfo(v, TPIType.TASKWAIT, ["taskwait"], v_first_line, [], [], [])
                    suggestions.append(tmp_suggestion)
            else:
                # no barrier needed
                pass
        elif omittable_count == 0 and task_count > 1:  # connected to at least two distinct task nodes
            if v.tp_contains_taskwait is False:
                # actual change
                v.tp_contains_taskwait = True
                barrier_nodes.append(v)
                transformation_happened = True
                tmp_suggestion = TaskParallelismInfo(v, TPIType.TASKWAIT, ["taskwait"], v_first_line, [], [], [])
                suggestions.append(tmp_suggestion)
        if omittable_count == 1 and v.tp_contains_task is False and v.tp_contains_taskwait is False:
            # omittable node appended to prior omittable node
            # get parent task
            #            parent_task: Optional[CUNode] = None
            for e in out_dep_edges:
                if pet.node_at(e[1]).tp_omittable is True:
                    # if tp_omittable is set, a omittable_suggestion has to exists.
                    # find this suggestion and extract combine_with_node
                    found_cwn = False
                    for tmp_omit, tmp_cwn in omittable_nodes:
                        if pet.node_at(e[1]) == tmp_omit:
                            if len(tmp_cwn) == 1:
                                parent_task = tmp_cwn[0]
                                found_cwn = True

                    if not found_cwn:
                        raise Exception("No parent task for omittable node found!")
            violation = __check_dependences_and_predecessors(pet, out_dep_edges, parent_task, v)
            # suggest omittable cu if no violation occured
            if not violation:
                if v.tp_omittable is False:
                    # actual change
                    v.tp_omittable = True
                    omittable_nodes.append((v, [parent_task]))
                    suggestions.append(OmittableCuInfo(v, parent_task))
                    transformation_happened = True

        # append neighbors of modified node to queue
        if transformation_happened:
            in_dep_edges = [
                (s, t, e) for s, t, e in pet.in_edges(v.id) if e.etype == EdgeType.DATA and pet.node_at(s) != v
            ]
            for e in out_dep_edges:
                queue.append(pet.node_at(e[1]))
            for e in in_dep_edges:
                queue.append(pet.node_at(e[0]))
            queue = list(set(queue))

    return suggestions


def __count_adjacent_nodes(
    pet: PEGraphX,
    suggestions: List[PatternInfo],
    out_dep_edges: List[Tuple[Any, Any, Any]],
    task_nodes: List[Node],
    barrier_nodes: List[Node],
    omittable_nodes: List[Tuple[Node, List[Node]]],
) -> Tuple[int, int, int, int]:
    """Checks the types of nodes pointed to by out_dep_edges and increments the respective counters.
    :param pet: PET Graph
    :param suggestions: List[TaskParallelismInfo]
    :param out_dep_edges: list of outgoing edges
    :param task_nodes: list of cu nodes containing task suggestions
    :param barrier_nodes: list of cu nodes containing barrier suggestions
    :param omittable_nodes: list of cu nodes containing omittable suggestions
    :return: Tuple consisting of (task_count, barrier_count, omittable_count, normal_count)
    """
    task_count = 0
    barrier_count = 0
    omittable_count = 0
    normal_count = 0
    task_buffer = []
    barrier_buffer = []
    omittable_parent_buffer = []
    for e in out_dep_edges:
        if pet.node_at(e[1]) in task_nodes:
            # only count distinct tasks
            if pet.node_at(e[1]) not in task_buffer:
                task_buffer.append(pet.node_at(e[1]))
                task_count += 1
            else:
                pass
        elif pet.node_at(e[1]) in barrier_nodes:
            # only count distinct barriers
            if pet.node_at(e[1]) not in barrier_buffer:
                barrier_buffer.append(pet.node_at(e[1]))
                barrier_count += 1
            else:
                pass
        elif pet.node_at(e[1]) in [tmp[0] for tmp in omittable_nodes]:
            # treat omittable cus like their parent tasks
            tmp_omit_suggestions: List[OmittableCuInfo] = cast(
                List[OmittableCuInfo], [s for s in suggestions if type(s) == OmittableCuInfo]
            )
            parent_task = [tos for tos in tmp_omit_suggestions if tos._node == pet.node_at(e[1])][0].combine_with_node
            if parent_task.id not in omittable_parent_buffer:
                omittable_parent_buffer.append(parent_task.id)
                omittable_count += 1
            else:
                pass
        else:
            normal_count += 1
    return task_count, barrier_count, omittable_count, normal_count


def __check_dependences_and_predecessors(
    pet: PEGraphX, out_dep_edges: List[Tuple[Any, Any, Any]], parent_task: Node, cur_cu: Node
):
    """Checks if only dependences to self, parent omittable node or path to target task exists.
    Checks if node is a direct successor of an omittable node or a task node.
    :param pet: PET Graph
    :param out_dep_edges: list of outgoing edges
    :param parent_task: parent cu of cur_cu
    :param cur_cu: current cu node
    :return True, if a violation has been found. False, otherwise.
    """
    violation = False
    # check if only dependencies to self, parent omittable node or path to target task exists
    for e in out_dep_edges:
        if pet.node_at(e[1]) == cur_cu:
            continue
        elif pet.node_at(e[1]).tp_omittable is True:
            continue
        elif check_reachability(pet, parent_task, cur_cu, [EdgeType.DATA]):
            continue
        else:
            violation = True
    # check if node is a direct successor of an omittable node or a task node
    in_succ_edges = [(s, t, e) for (s, t, e) in pet.in_edges(cur_cu.id) if e.etype == EdgeType.SUCCESSOR]
    is_successor = False
    for e in in_succ_edges:
        if pet.node_at(e[0]).tp_omittable is True:
            is_successor = True
        elif pet.node_at(e[0]).tp_contains_task is True:
            is_successor = True
    if not is_successor:
        violation = True
    return violation


def __split_suggestions(
    suggestions: List[PatternInfo],
    taskwait_suggestions: List[TaskParallelismInfo],
    task_suggestions: List[TaskParallelismInfo],
    omittable_suggestions: List[PatternInfo],
):
    """Split suggestions into taskwait, task and omittable suggestions.
    :param suggestions: list of suggestions to be split
    :param taskwait_suggestions: list to store taskwait suggestions
    :param task_suggestions: list to store task suggestions
    :param omittable_suggestions: list to store omittable suggestions"""
    for single_suggestion in suggestions:
        if type(single_suggestion) == ParallelRegionInfo:
            continue
        elif type(single_suggestion) == OmittableCuInfo:
            omittable_suggestions.append(single_suggestion)
        elif type(single_suggestion) == TaskParallelismInfo:
            if single_suggestion.type is TPIType.TASKWAIT:
                taskwait_suggestions.append(single_suggestion)
            elif single_suggestion.type is TPIType.TASK:
                task_suggestions.append(single_suggestion)
        else:
            raise TypeError("Unknown Type: ", type(single_suggestion))


def suggest_barriers_for_uncovered_tasks_before_return(
    pet: PEGraphX, suggestions: List[PatternInfo]
) -> List[PatternInfo]:
    """enforces taskwait or similar pragmas before return statements to ensure, that no unfinished tasks exist
    when the parent function returns.
    :param pet: PET graph
    :param suggestions; List[PatternInfo]
    :return: List[PatternInfo]"""
    # iterate over task suggestions
    for suggestion in suggestions:
        if type(suggestion) != TaskParallelismInfo:
            continue
        if suggestion.type is not TPIType.TASK:
            continue
        # if task is covered by a parallel region, ignore it due to the present, implicit barrier
        covered_by_parallel_region = False
        for tmp in suggestions:
            if type(tmp) == ParallelRegionInfo:
                if line_contained_in_region(suggestion.start_line, tmp.region_start_line, tmp.region_end_line):
                    covered_by_parallel_region = True
                    break
        if covered_by_parallel_region:
            continue
        # check, if barrier in successor - path between task and return (same cu -> no barrier contained)
        queue = [suggestion._node]
        visited = []
        targets = []
        while len(queue) != 0:
            current_node = queue.pop()
            visited.append(current_node)
            if current_node.tp_contains_taskwait:
                # stop search on this path
                continue
            # check if returnInstructionCount > 0
            if isinstance(current_node, CUNode) and current_node.return_instructions_count > 0:
                # taskwait missing -> add current node to targets
                targets.append(current_node)
                continue
            # append direct successors to targets, if not in visited
            successors = pet.direct_successors(current_node)
            successors = [ds for ds in successors if ds not in visited]
            queue = queue + successors
        # suggest taskwait prior to return if needed
        for cu in targets:
            # actual change
            cu.tp_contains_taskwait = True
            pragma_line = cu.end_position()  # since return has to be the last statement in a CU
            pragma_line = LineID(pragma_line[pragma_line.index(":") + 1 :])
            tmp_suggestion = TaskParallelismInfo(cu, TPIType.TASKWAIT, ["taskwait"], pragma_line, [], [], [])
            print(
                "TPDet:suggest_barriers_for_uncovered_tasks_before_return: added taskwait suggestion at line: ",
                cu.end_position(),
            )
            suggestions.append(tmp_suggestion)
    return suggestions


def validate_barriers(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """Checks if >= 2 dependencies exist from same successor path or
    node that contains the barrier is of type loop.
    Eliminate those barrier suggestions that violate this requirement.
    A successor path is represented by a list of nodes reachable by traversing
    the successor edges inside a single function in reverse direction.
    Note, that nodes with multiple outgoing successor edges
    (multiple control flow options) lead to a separation of the created
    successor paths to support the desired behavior.
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :return: List[PatternInfo]
    """
    barrier_suggestions: List[TaskParallelismInfo] = []
    result: List[PatternInfo] = []
    for single_suggestion in suggestions:
        if type(single_suggestion) == TaskParallelismInfo:
            try:
                if single_suggestion.type is TPIType.TASKWAIT:
                    barrier_suggestions.append(single_suggestion)
                else:
                    result.append(single_suggestion)
            except AttributeError:
                result.append(single_suggestion)
        else:
            result.append(single_suggestion)

    for bs in barrier_suggestions:
        # check if type of bs node is loop and accept the suggestion if so
        # reason: if task is spawned inside a loop, paths are irrelevant
        if bs._node.type == NodeType.LOOP:
            result.append(bs)
            continue

        # create "path lists" for each incoming successor edge
        in_succ_edges = [
            (s, t, e)
            for s, t, e in pet.in_edges(bs._node.id)
            if e.etype == EdgeType.SUCCESSOR and pet.node_at(s) != bs._node
        ]
        predecessors_dict = dict()
        for e in in_succ_edges:
            visited_nodes: List[Node] = []
            tmp, visited_nodes = get_predecessor_nodes(pet, pet.node_at(e[0]), visited_nodes)
            predecessors_dict[e] = tmp
        # iterate over outgoing dependence edges and increase dependence counts
        # for those paths that contain the dependence target CU
        out_dep_edges = [
            (s, t, e)
            for s, t, e in pet.out_edges(bs._node.id)
            if e.etype == EdgeType.DATA and pet.node_at(t) != bs._node
        ]
        dependence_count_dict = dict()

        for key in predecessors_dict:
            dependence_count_dict[key] = 0

        for key in predecessors_dict:
            for e in out_dep_edges:
                if pet.node_at(e[1]) in predecessors_dict[key]:
                    dependence_count_dict[key] += 1

        # if validated, append bs to result
        validation_successful = False
        for key in dependence_count_dict:
            if dependence_count_dict[key] > 1:
                result.append(bs)
                validation_successful = True
                break
        # if not validated, unmark node as containing a taskwait in the graph
        if not validation_successful:
            bs._node.tp_contains_taskwait = False

    return result


def suggest_missing_barriers_for_global_vars(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """Suggests a barrier if a node is a successor of a task CU
    which is not covered by an existing barrier and the set of global variables
    of the CU and the task are overlapping
    (i.e. both CUs access common global variables).
    If the cu which would be suggested as a barrier contains a Task suggestion
    already, ignore the barrier suggestion
    (reason: false positives due to copying of global / local variables in preprocessor).
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :return: List[PatternInfo]
    """
    # split suggestions into task and taskwait suggestions
    taskwait_suggestions = []
    task_suggestions = []
    for single_suggestion in suggestions:
        if type(single_suggestion) == ParallelRegionInfo or type(single_suggestion) == OmittableCuInfo:
            continue
        if type(single_suggestion) == TaskParallelismInfo:
            if single_suggestion.type is TPIType.TASKWAIT:
                taskwait_suggestions.append(single_suggestion)
            elif single_suggestion.type is TPIType.TASK:
                task_suggestions.append(single_suggestion)
        else:
            raise TypeError("Unsupported Type: ", type(single_suggestion))

    # iterate over task suggestions
    for task_sug in task_suggestions:
        visited_nodes = [task_sug._node]
        out_succ_edges = [
            (s, t, e)
            for s, t, e in pet.out_edges(task_sug._node.id)
            if e.etype == EdgeType.SUCCESSOR and pet.node_at(t) != task_sug._node
        ]
        queue = out_succ_edges
        # iterate over queued successor-edges
        while len(queue) > 0:
            succ_edge = queue.pop()
            if not pet.node_at(succ_edge[1]) in visited_nodes:
                visited_nodes.append(pet.node_at(succ_edge[1]))
            else:
                continue
            # if barrier is encountered, stop
            if pet.node_at(succ_edge[1]).tp_contains_taskwait is True:
                continue
            # if edge.target has common global variable with task
            succ = pet.node_at(succ_edge[1])
            common_vars = [
                var
                for var in (succ.global_vars if isinstance(succ, CUNode) else [])
                if var in cast(CUNode, task_sug._node).global_vars
            ]
            if len(common_vars) > 0:
                # if cu is a task suggestion, continue
                if pet.node_at(succ_edge[1]).tp_contains_task is True:
                    continue
                # check if any element of common vars is not contained in task_sug.out_dep
                if len([v for v in task_sug.out_dep if v not in [e.name for e in common_vars]]) > 0:
                    # suggest taskwait
                    if pet.node_at(succ_edge[1]).tp_contains_taskwait is False:
                        # actual change
                        pet.node_at(succ_edge[1]).tp_contains_taskwait = True
                        first_line = pet.node_at(succ_edge[1]).start_position()
                        first_line = LineID(first_line[first_line.index(":") + 1 :])
                        tmp_suggestion = TaskParallelismInfo(
                            pet.node_at(succ_edge[1]),
                            TPIType.TASKWAIT,
                            ["taskwait"],
                            first_line,
                            [],
                            [],
                            [],
                        )
                        suggestions.append(tmp_suggestion)
                    continue
            # append current nodes outgoing successor edges to queue
            target_out_succ_edges = [
                (s, t, e)
                for s, t, e in pet.out_edges(pet.node_at(succ_edge[1]).id)
                if e.etype == EdgeType.SUCCESSOR and pet.node_at(t) != pet.node_at(succ_edge[1])
            ]
            queue = list(set(queue + target_out_succ_edges))
    return suggestions



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from argparse import ArgumentParser

from discopop_library.discopop_optimizer.optimizer import run
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments


def parse_args() -> OptimizerArguments:
    """Parse the arguments passed to the discopop_optimizer"""
    parser = ArgumentParser(description="DiscoPoP Optimizer")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments marked as experimental features. These flags may or may not be removed or changed in the future.",
    )

    # fmt: off
    parser.add_argument("-v", "--verbose", action="store_true",
        help="Enable verbose output.")
    parser.add_argument("-x", "--exhaustive", action="store_true",
        help="Enable exhaustive search. By default, an evolutionary search with a population size of 50 and 5 generations is performed.")
    parser.add_argument("-e", "--evolutionary", type=str, default=None, nargs=2, metavar=("population_size", "generations"),
        help="Enable evolutionary search. By default, an evolutionary search with a population size of 50 and 5 generations is performed.")
    parser.add_argument(
        "--doall-microbench-file", type=str, default="None",
        help="Do-All microbenchmark results"
    )
    parser.add_argument(
        "--reduction-microbench-file", type=str, default="None",
        help="Reduction microbenchmark results"
    )
    parser.add_argument(
        "--system-configuration", type=str, default="optimizer/system_configuration.json",
        help="System configuration file"
    )
    # EXPERIMENTAL FLAGS:
    experimental_parser.add_argument("--allow-nested-parallelism", action="store_true",
        help="Allow the creation of nested parallelism suggestions. "
        + "WARNING: Cost estimations may not be accurrate due to potentially"
        + "high overhead introduced by entering nested parallelism!")
    experimental_parser.add_argument("--check-called-function-for-nested-parallelism", action="store_true", help="Extend the check for nested parallelism to called functions."
        + "WARNING: Execution time may increase significantly!")
    experimental_parser.add_argument("-i", "--interactive", action="store_true",
        help="Enable interactive execution.")
    experimental_parser.add_argument("--plot", action="store_true",
        help="Plot the internal graph.")
    # fmt: on

    arguments = parser.parse_args()

    return OptimizerArguments(
        verbose=arguments.verbose,
        interactive=arguments.interactive,
        exhaustive=arguments.exhaustive,
        evolutionary=arguments.evolutionary,
        doall_microbench_file=arguments.doall_microbench_file,
        reduction_microbench_file=arguments.reduction_microbench_file,
        allow_nested_parallelism=arguments.allow_nested_parallelism,
        plot=arguments.plot,
        system_configuration_path=arguments.system_configuration,
        check_called_function_for_nested_parallelism=arguments.check_called_function_for_nested_parallelism,
    )


def main():
    arguments = parse_args()
    run(arguments)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.suggestions.optimizers.loop_collapse import collapse_loops

import networkx as nx  # type: ignore


def optimize_suggestions(experiment: Experiment) -> nx.DiGraph:
    experiment.optimization_graph = collapse_loops(experiment)
    return experiment.optimization_graph



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import argparse
import os
from os.path import dirname

from DEPRECATED_discopop_wizard.classes.Arguments import Arguments
from DEPRECATED_discopop_wizard.headless.headless_execution import (
    execute_all_stored_configurations,
    execute_tag_filtered_configurations,
)
from DEPRECATED_discopop_wizard.wizard import main as wizard_main


def convert_args(namespace_args: argparse.Namespace) -> Arguments:
    """Stores args into an Arguments object."""
    return_arg = Arguments()
    return_arg.execute_configurations_with_tag = [
        tag for tag in namespace_args.execute_configurations_with_tag.split(",") if len(tag) > 0
    ]
    return_arg.execute_all_configurations = namespace_args.execute_all_configurations == "true"

    return return_arg


def main():
    parser = argparse.ArgumentParser(description="DiscoPoP Configuration Wizard")

    parser.add_argument(
        "--execute_all_configurations",
        help="Execute all stored configurations in a headless manner. [true / false]",
    )
    parser.add_argument(
        "--execute_configurations_with_tag",
        default="",
        help="Execute all stored configurations in a headless manner which have any of the given tags assigned. [comma-separated list of tags to be executed]",
    )

    args = convert_args(parser.parse_args())

    if args.execute_all_configurations:
        # start headless mode and execute all stored configurations
        source_dir = dirname(os.path.abspath(__file__))  # source_dir: discopop/discopop_wizard
        execute_all_stored_configurations(args, source_dir)
    elif len(args.execute_configurations_with_tag) > 0:
        # start headless mode and execute stored configurations with the suitable tags
        source_dir = dirname(os.path.abspath(__file__))  # source_dir: discopop/discopop_wizard
        execute_tag_filtered_configurations(args, source_dir)
    else:
        wizard_main(args)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os.path
from pathlib import Path
from typing import Dict

from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments
from discopop_library.PatchApplicator.apply import apply_patches


def load_patches(
    file_mapping: Dict[int, Path],
    arguments: PatchApplicatorArguments,
    applied_suggestions_file: str,
    patch_generator_dir: str,
) -> int:
    # check if a saved configuration exists
    if not os.path.exists(applied_suggestions_file + ".save"):
        raise FileNotFoundError(
            "Nothing to be loaded. Exiting." "\nExpected file: ", applied_suggestions_file + ".save"
        )

    # get the list of suggestions to be applied
    with open(applied_suggestions_file + ".save", "r") as f:
        suggestions_to_be_applied = json.loads(f.read())["applied"]

    # apply all suggestions
    retval = apply_patches(
        suggestions_to_be_applied, file_mapping, arguments, applied_suggestions_file, patch_generator_dir
    )

    return retval



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from multiprocessing import Pool
from typing import List, Dict, Set, Tuple, cast

from alive_progress import alive_bar  # type: ignore

from .PatternInfo import PatternInfo
from ..PEGraphX import (
    CUNode,
    LoopNode,
    PEGraphX,
    Node,
    NodeType,
    EdgeType,
    LineID,
    MemoryRegion,
    DepType,
)
from ..utils import classify_loop_variables
from ..variable import Variable


class DoAllInfo(PatternInfo):
    """Class, that contains do-all detection result"""

    def __init__(self, pet: PEGraphX, node: Node):
        """
        :param pet: PET graph
        :param node: node, where do-all was detected
        """
        PatternInfo.__init__(self, node)
        fp, p, lp, s, r = classify_loop_variables(pet, node)
        self.first_private = fp
        self.private = p
        self.last_private = lp
        self.shared = s
        self.reduction = r
        self.scheduling_clause = "static"
        self.collapse_level = 1

    def __str__(self):
        return (
            f"Do-all at: {self.node_id}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            # f"iterations: {self.iterations_count}\n"
            # f"instructions: {self.instructions_count}\n"
            # f"workload: {self.workload}\n"
            f'pragma: "#pragma omp parallel for"\n'
            f"collapse: {self.collapse_level}\n"
            f"private: {[v.name for v in self.private]}\n"
            f"shared: {[v.name for v in self.shared]}\n"
            f"first private: {[v.name for v in self.first_private]}\n"
            f"reduction: {[v.name for v in self.reduction]}\n"
            f"last private: {[v.name for v in self.last_private]}\n"
            f"scheduling clause: {self.scheduling_clause}"
        )


global_pet = None


def run_detection(pet: PEGraphX) -> List[DoAllInfo]:
    """Search for do-all loop pattern

    :param pet: PET graph
    :return: List of detected pattern info
    """
    import tqdm  # type: ignore

    global global_pet
    global_pet = pet
    result: List[DoAllInfo] = []
    nodes = pet.all_nodes(LoopNode)

    param_list = [(node) for node in nodes]
    with Pool(initializer=__initialize_worker, initargs=(pet,)) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__check_node, param_list), total=len(param_list)))
    for local_result in tmp_result:
        result += local_result
    print("GLOBAL RES: ", result)

    for pattern in result:
        pattern.get_workload(pet)
        pattern.get_per_iteration_workload(pet)

    # remove reduction operations from shared variables to prevent issues / incorrect results in the exported JSON file
    for idx, _ in enumerate(result):
        for idx_2, _ in enumerate(result[idx].shared):
            result[idx].shared[idx_2].operation = None

    return result


def __initialize_worker(pet):
    global global_pet
    global_pet = pet


def __check_node(param_tuple):
    global global_pet
    local_result = []
    node = param_tuple

    if __detect_do_all(global_pet, node):
        node.do_all = True
        if not node.reduction and node.loop_iterations >= 0 and not node.contains_array_reduction:
            local_result.append(DoAllInfo(global_pet, node))

    return local_result


def __detect_do_all(pet: PEGraphX, root_loop: LoopNode) -> bool:
    """Calculate do-all value for node

    :param pet: PET graph
    :param root: root node
    :return: true if do-all
    """
    subnodes = [pet.node_at(t) for s, t, d in pet.out_edges(root_loop.id, [EdgeType.CHILD, EdgeType.CALLSNODE])]

    # get required metadata
    loop_start_lines: List[LineID] = []
    root_children = pet.subtree_of_type(root_loop, (CUNode, LoopNode))
    root_children_cus = [cast(CUNode, cu) for cu in root_children if cu.type == NodeType.CU]
    root_children_loops = [cast(LoopNode, cu) for cu in root_children if cu.type == NodeType.LOOP]
    for v in root_children_loops:
        loop_start_lines.append(v.start_position())
    fp, p, lp, s, r = classify_loop_variables(pet, root_loop)
    # get variables which are defined inside the loop
    defined_inside_loop: List[Tuple[Variable, Set[MemoryRegion]]] = []
    tmp_loop_variables = pet.get_variables(root_children_cus)
    for var in tmp_loop_variables:
        if var.defLine >= root_loop.start_position() and var.defLine <= root_loop.end_position():
            defined_inside_loop.append((var, tmp_loop_variables[var]))

    # check if all subnodes are parallelizable
    for node in pet.subtree_of_type(root_loop, CUNode):
        if node.performs_file_io:
            # node is not reliably parallelizable as some kind of file-io is performed.
            return False

    for i in range(0, len(subnodes)):
        children_cache: Dict[Node, List[Node]] = dict()
        dependency_cache: Dict[Tuple[Node, Node], Set[Node]] = dict()
        for j in range(i, len(subnodes)):
            if __check_loop_dependencies(
                pet,
                subnodes[i],
                subnodes[j],
                root_loop,
                root_children_cus,
                root_children_loops,
                loop_start_lines,
                fp,
                p,
                lp,
                defined_inside_loop,
            ):
                # if pet.depends_ignore_readonly(subnodes[i], subnodes[j], root_loop):
                return False

    return True


def __check_loop_dependencies(
    pet: PEGraphX,
    node_1: Node,
    node_2: Node,
    root_loop: LoopNode,
    root_children_cus: List[CUNode],
    root_children_loops: List[LoopNode],
    loop_start_lines: List[LineID],
    first_privates: List[Variable],
    privates: List[Variable],
    last_privates: List[Variable],
    defined_inside_loop: List[Tuple[Variable, Set[MemoryRegion]]],
) -> bool:
    """Returns True, if dependencies between the respective subgraphs chave been found.
    Returns False otherwise, which results in the potential suggestion of a Do-All pattern."""
    # get recursive children of source and target
    node_1_children_ids = [node.id for node in pet.subtree_of_type(node_1, CUNode)]
    node_2_children_ids = [node.id for node in pet.subtree_of_type(node_2, CUNode)]

    # get dependency edges between children nodes
    deps = set()
    for n in node_1_children_ids + node_2_children_ids:
        deps.update(
            [(s, t, d) for s, t, d in pet.in_edges(n, EdgeType.DATA) if s in node_1_children_ids + node_2_children_ids]
        )
        deps.update(
            [(s, t, d) for s, t, d in pet.out_edges(n, EdgeType.DATA) if t in node_1_children_ids + node_2_children_ids]
        )

    # get memory regions which are defined inside the loop
    memory_regions_defined_in_loop = set()
    for var, mem_regs in defined_inside_loop:
        memory_regions_defined_in_loop.update(mem_regs)

    for source, target, dep in deps:
        # check if targeted variable is readonly inside loop
        if pet.is_readonly_inside_loop_body(
            dep,
            root_loop,
            root_children_cus,
            root_children_loops,
            loops_start_lines=loop_start_lines,
        ):
            # variable is readonly -> no problem
            continue

        # check if targeted variable is loop index
        if pet.is_loop_index(dep.var_name, loop_start_lines, root_children_cus):
            continue

        # check if variable is defined inside loop
        if dep.memory_region in memory_regions_defined_in_loop:
            continue

        # targeted variable is not read-only
        if dep.dtype == DepType.INIT:
            continue
        elif dep.dtype == DepType.RAW:
            # check RAW dependencies
            # RAW problematic, if it is not an intra-iteration RAW
            if not dep.intra_iteration:
                return True
            # if it is an intra iteration dependency, it is problematic if it belongs to a parent loop
            elif dep.intra_iteration_level > root_loop.get_nesting_level(pet):
                tmp = root_loop.get_nesting_level(pet)
                return True

        elif dep.dtype == DepType.WAR:
            # check WAR dependencies
            # WAR problematic, if it is not an intra-iteration WAR and the variable is not private or firstprivate
            if not dep.intra_iteration:
                if dep.var_name not in [v.name for v in first_privates + privates + last_privates]:
                    return True
            # if it is an intra iteration dependency, it is problematic if it belongs to a parent loop
            elif dep.intra_iteration_level > root_loop.get_nesting_level(pet):
                tmp = root_loop.get_nesting_level(pet)
                return True
        elif dep.dtype == DepType.WAW:
            # check WAW dependencies
            # handled by variable classification
            pass
        else:
            raise ValueError("Unsupported dependency type: ", dep.dtype)

    # no problem found. Potentially suggest Do-All
    return False


def __old_detect_do_all(pet: PEGraphX, root_loop: CUNode) -> bool:
    """Calculate do-all value for node

    :param pet: PET graph
    :param root: root node
    :return: true if do-all
    """
    subnodes = [pet.node_at(t) for s, t, d in pet.out_edges(root_loop.id, [EdgeType.CHILD, EdgeType.CALLSNODE])]

    # check if all subnodes are parallelizable
    for node in pet.subtree_of_type(root_loop, CUNode):
        if node.performs_file_io:
            # node is not reliably parallelizable as some kind of file-io is performed.
            return False

    for i in range(0, len(subnodes)):
        children_cache: Dict[CUNode, List[CUNode]] = dict()
        dependency_cache: Dict[Tuple[CUNode, CUNode], Set[CUNode]] = dict()
        for j in range(i, len(subnodes)):
            if pet.depends_ignore_readonly(subnodes[i], subnodes[j], root_loop):
                return False

    return True



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import subprocess
from typing import Union, cast, Dict, List, Tuple, Optional, Any

from lxml import objectify  # type: ignore

from discopop_explorer.PEGraphX import (
    DummyNode,
    FunctionNode,
    CUNode,
    Node,
    NodeType,
    EdgeType,
    MWType,
    DepType,
    PEGraphX,
    LineID,
)
from discopop_explorer.pattern_detectors.task_parallelism.classes import Task, TaskParallelismInfo
from discopop_explorer.utils import depends

__workloadThreshold = 10000
__minParallelism = 3
__forks = set()
__global_llvm_cxxfilt_path: str = ""
demangling_cache: Dict[str, str] = dict()


def demangle(mangled_name: str) -> str:
    """Demangles the given mangled function name and returns the resulting string after demangling.
    :param mangled_name: mangled function name
    :return: demangled function name and type information"""
    global demangling_cache
    global __global_llvm_cxxfilt_path
    if mangled_name in demangling_cache:
        return demangling_cache[mangled_name]
    if __global_llvm_cxxfilt_path == "None":
        # set default llvm-cxxfilt executable
        llvm_cxxfilt_path = "llvm-cxxfilt"
    else:
        llvm_cxxfilt_path = __global_llvm_cxxfilt_path
    try:
        process = subprocess.Popen([llvm_cxxfilt_path, mangled_name], stdout=subprocess.PIPE)
        process.wait()
        if process.stdout is not None:
            out_bytes = process.stdout.readline()
            out = out_bytes.decode("UTF-8")
            out = out.replace("\n", "")
            demangling_cache[mangled_name] = out
            return out
    except FileNotFoundError:
        raise ValueError(
            "Executable '" + llvm_cxxfilt_path + "' not found." + " Check or supply --llvm-cxxfilt-path parameter."
        )
    raise ValueError("Demangling of " + mangled_name + " not possible!")


def line_contained_in_region(test_line: str, start_line: str, end_line: str) -> bool:
    """check if test_line is contained in [startLine, endLine].
    Return True if so. False else.
    :param test_line: <fileID>:<line>
    :param start_line: <fileID>:<line>
    :param end_line: <fileID>:<line>
    :return: bool
    """
    test_line_file_id = int(test_line.split(":")[0])
    test_line_line = int(test_line.split(":")[1])
    start_line_file_id = int(start_line.split(":")[0])
    start_line_line = int(start_line.split(":")[1])
    end_line_file_id = int(end_line.split(":")[0])
    end_line_line = int(end_line.split(":")[1])
    if (
        test_line_file_id == start_line_file_id == end_line_file_id
        and start_line_line <= test_line_line <= end_line_line
    ):
        return True
    return False


def get_parent_of_type(
    pet: PEGraphX, node: Node, parent_type: NodeType, edge_type: EdgeType, only_first: bool
) -> List[Tuple[Node, Optional[Node]]]:
    """return parent cu nodes and the last node of the path to them as a tuple
    for the given node with type parent_type
    accessible via edges of type edge_type.
    :param pet: PET graph
    :param node: CUNode, root for the search
    :param parent_type: NodeType, type of target node
    :param edge_type: EdgeType, type of usable edges
    :param only_first: Bool, if true, return only first parent.
        Else, return first parent for each incoming edge of node.
    :return: [(CUNode, CUNode)]"""
    visited = []
    queue: List[Tuple[Node, Optional[Node]]] = [(node, None)]
    res: List[Tuple[Node, Optional[Node]]] = []
    while len(queue) > 0:
        tmp = queue.pop(0)
        (cur_node, last_node) = tmp
        last_node = cur_node
        visited.append(cur_node)
        tmp_list = [
            (s, t, e) for s, t, e in pet.in_edges(cur_node.id) if pet.node_at(s) not in visited and e.etype == edge_type
        ]
        for e in tmp_list:
            if pet.node_at(e[0]).type == parent_type:
                if only_first is True:
                    return [(pet.node_at(e[0]), last_node)]
                else:
                    res.append((pet.node_at(e[0]), last_node))
                    visited.append(pet.node_at(e[0]))
            else:
                if pet.node_at(e[0]) not in visited:
                    queue.append((pet.node_at(e[0]), last_node))
    return res


def get_cus_inside_function(pet: PEGraphX, function_cu: Node) -> List[Node]:
    """Returns cus contained in function-body as a list.
    :param pet: PET Graph
    :param function_cu: target function node
    :return: List[CUNode]"""
    queue: List[Node] = [function_cu]
    result_list: List[Node] = []
    while len(queue) > 0:
        cur_cu = queue.pop(0)
        if cur_cu not in result_list:
            result_list.append(cur_cu)
        # append children to queue
        for e in pet.out_edges(cur_cu.id, [EdgeType.CHILD]):
            queue.append(pet.node_at(e[1]))
    return result_list


def check_reachability(pet: PEGraphX, target: Node, source: Node, edge_types: List[EdgeType]) -> bool:
    """check if target is reachable from source via edges of types edge_type.
    :param pet: PET graph
    :param source: CUNode
    :param target: CUNode
    :param edge_types: List[EdgeType]
    :return: Boolean"""
    if source == target:
        return True
    visited: List[str] = []
    queue = [target]
    while len(queue) > 0:
        cur_node = queue.pop(0)
        visited.append(cur_node.id)
        tmp_list = [(s, t, e) for s, t, e in pet.in_edges(cur_node.id) if s not in visited and e.etype in edge_types]
        for e in tmp_list:
            if pet.node_at(e[0]) == source:
                return True
            else:
                if e[0] not in visited:
                    queue.append(pet.node_at(e[0]))
    return False


def get_predecessor_nodes(pet: PEGraphX, root: Node, visited_nodes: List[Node]) -> Tuple[List[Node], List[Node]]:
    """return a list of reachable predecessor nodes.
    generate list recursively.
    stop recursion if a node of type "function" is found or root is a barrier
    (predecessing barrier of the original root node, further predecessors are
    already covered by this barrier and thus can be ignored).
    :param pet: PET Graph
    :param root: root node of the search
    :param visited_nodes: list of visited nodes
    :return: Tuple[[predecessor nodes], [visited nodes]]"""
    result = [root]
    visited_nodes.append(root)
    if isinstance(root, FunctionNode) or root.tp_contains_taskwait is True:
        # root of type "function" or root is a barrier
        return result, visited_nodes
    in_succ_edges = [
        (s, t, e)
        for s, t, e in pet.in_edges(root.id)
        if e.etype == EdgeType.SUCCESSOR and pet.node_at(s) != root and pet.node_at(s) not in visited_nodes
    ]
    for e in in_succ_edges:
        tmp, visited_nodes = get_predecessor_nodes(pet, pet.node_at(e[0]), visited_nodes)
        result += tmp
    return result, visited_nodes


def check_neighbours(first: Task, second: Task):
    """Checks if second task immediately follows first task

    :param first: predecessor task
    :param second: successor task
    :return: true if second task immediately follows first task
    """
    fel = int(first.end_line.split(":")[1])
    ssl = int(second.start_line.split(":")[1])
    return fel == ssl or fel + 1 == ssl or fel + 2 == ssl


# NOTE: left, as it may become relevant in the near future
# We decided to omit the information that computes the workload and the relevant codes. For large programs (e.g., ffmpeg), the generated Data.xml file becomes very large. However, we keep the code here because we would like to integrate a hotspot detection algorithm (TODO: Bertin) with the parallelism discovery. Then, we need to retrieve the information to decide which code sections (loops or functions) are worth parallelizing.
# def merge_tasks(pet: PETGraphX, task: Task):
#    """Merges the tasks into having required workload.
#
#    :param pet: PET graph
#    :param task: task node
#    """
#    for i in range(len(task.child_tasks)):
#        child_task: Task = task.child_tasks[i]
#        if child_task.workload < __workloadThreshold:
#            if i > 0:
#                pred: Task = task.child_tasks[i - 1]
#                if check_neighbours(pred, child_task):
#                    pred.aggregate(child_task)
#                    pred.child_tasks.remove(child_task)
#                    merge_tasks(pet, task)
#                    return
#            if i + 1 < len(task.child_tasks) - 1:
#                succ: Task = task.child_tasks[i + 1]
#                if check_neighbours(child_task, succ):
#                    child_task.aggregate(succ)
#                    task.child_tasks.remove(succ)
#                    merge_tasks(pet, task)
#                    return
#            task.child_tasks.remove(child_task)
#            merge_tasks(pet, task)
#            return
#
#    if task.child_tasks and len(task.child_tasks) < __minParallelism:
#        max_workload_task = max(task.child_tasks, key=lambda t: t.workload)
#        task.child_tasks.extend(max_workload_task.child_tasks)
#        task.child_tasks.remove(max_workload_task)
#        merge_tasks(pet, task)
#        return
#
#    for child in task.child_tasks:
#        if child.nodes[0].type == NodeType.LOOP:
#            pass


def create_task_tree(pet: PEGraphX, root: Node):
    """generates task tree data from root node

    :param pet: PET graph
    :param root: root node
    """
    root_task = Task(pet, root)
    __forks.add(root_task)
    create_task_tree_helper(pet, root, root_task, [])


def create_task_tree_helper(pet: PEGraphX, current: Node, root: Task, visited_func: List[Node]):
    """generates task tree data recursively

    :param pet: PET graph
    :param current: current vertex to process
    :param root: root task for subtree
    :param visited_func: visited function nodes
    """
    if isinstance(current, FunctionNode):
        if current in visited_func:
            return
        else:
            visited_func.append(current)

    for child in pet.direct_children_or_called_nodes(current):
        mw_type = child.mw_type

        if mw_type in [MWType.BARRIER, MWType.BARRIER_WORKER, MWType.WORKER]:
            task = Task(pet, child)
            root.child_tasks.append(task)
            create_task_tree_helper(pet, child, task, visited_func)
        elif mw_type == MWType.FORK and not child.start_position().endswith("16383"):
            task = Task(pet, child)
            __forks.add(task)
            create_task_tree_helper(pet, child, task, visited_func)
        else:
            create_task_tree_helper(pet, child, root, visited_func)


def recursive_function_call_contained_in_worker_cu(function_call_string: str, worker_cus: List[Node]) -> Node:
    """check if submitted function call is contained in at least one WORKER cu.
    Returns the vertex identifier of the containing cu.
    If no cu contains the function call, None is returned.
    Note: The Strings stored in recursiveFunctionCalls might contain multiple function calls at once.
          in order to apply this function correctly, make sure to split Strings in advance and supply
          one call at a time.
    :param function_call_string: String representation of the recursive function call to be checked
            Ex.: fib 7:35,  (might contain ,)
    :param worker_cus: List of vertices
    :return: CUNode
    """
    # remove , and whitespaces at start / end
    function_call_string = function_call_string.replace(",", "")
    while function_call_string.startswith(" "):
        function_call_string = function_call_string[1:]
    while function_call_string.endswith(" "):
        function_call_string = function_call_string[:-1]
    # function_call_string looks now like like: 'fib 7:52'

    # split String into function_name. file_id and line_number
    file_id = function_call_string[function_call_string.index(" ") + 1 : function_call_string.index(":")]
    line_number = function_call_string[function_call_string.index(":") + 1 :]

    # get tightest surrounding cu
    tightest_worker_cu = None
    # iterate over worker_cus
    for cur_w in worker_cus:
        cur_w_starts_at_line = cur_w.start_position()
        cur_w_ends_at_line = cur_w.end_position()
        cur_w_file_id = cur_w_starts_at_line[: cur_w_starts_at_line.index(":")]
        # check if file_id is equal
        if file_id == cur_w_file_id:
            # trim to line numbers only
            cur_w_starts_at_line = LineID(cur_w_starts_at_line[cur_w_starts_at_line.index(":") + 1 :])
            cur_w_ends_at_line = LineID(cur_w_ends_at_line[cur_w_ends_at_line.index(":") + 1 :])
            # check if line_number is contained
            if int(cur_w_starts_at_line) <= int(line_number) <= int(cur_w_ends_at_line):
                # check if cur_w is tighter than last result
                if tightest_worker_cu is None:
                    tightest_worker_cu = cur_w
                    continue
                if line_contained_in_region(
                    cur_w.start_position(),
                    tightest_worker_cu.start_position(),
                    tightest_worker_cu.end_position(),
                ) and line_contained_in_region(
                    cur_w.end_position(),
                    tightest_worker_cu.start_position(),
                    tightest_worker_cu.end_position(),
                ):
                    tightest_worker_cu = cur_w
    if tightest_worker_cu is None:
        raise ValueError("No surrounding worker CU could be found.")
    return tightest_worker_cu


def task_contained_in_reduction_loop(
    pet: PEGraphX, task: TaskParallelismInfo
) -> Tuple[Optional[Dict[str, str]], Optional[Node]]:
    """detect if task is contained in loop body of a reduction loop.
    return None, if task is not contained in reduction loop.
    else, return reduction_vars entry of parent reduction loop and loop CU Node.
    :param pet: PET graph
    :param task: TaskParallelismInfo
    :return: None / ({loop_line, name, reduction_line, operation}, CUNode)
    """
    # check if task contained in loop body
    parents = get_parent_of_type(pet, task._node, NodeType.LOOP, EdgeType.CHILD, False)
    contained_in = []
    if len(parents) == 0:
        return None, None
    else:
        # check if task is actually contained in one of the parents
        for parent_loop, last_node in parents:
            p_start_line = parent_loop.start_position()
            p_start_line = LineID(p_start_line[p_start_line.index(":") + 1 :])
            p_end_line = parent_loop.end_position()
            p_end_line = LineID(p_end_line[p_end_line.index(":") + 1 :])
            t_start_line = task.start_line
            t_start_line = LineID(t_start_line[t_start_line.index(":") + 1 :])
            t_end_line = task.end_line
            t_end_line = LineID(t_end_line[t_end_line.index(":") + 1 :])
            if p_start_line <= t_start_line and p_end_line >= t_end_line:
                contained_in.append(parent_loop)
    # check if task is contained in a reduction loop
    for parent in contained_in:
        if parent.reduction:
            # get correct entry for loop from pet.reduction_vars
            for rv in pet.reduction_vars:
                if rv["loop_line"] == parent.start_position():
                    return rv, parent
    return None, None


def get_function_call_from_source_code(
    source_code_files: Dict[str, str],
    line_number: int,
    file_id: str,
    called_function_name: Optional[str] = None,
) -> str:
    """Extract code snippet from original source code which contains a function call.
    :param source_code_files: File-Mapping dictionary
    :param line_number: original source code line number to start searching at
    :param file_id: file id of the original source code file
    :param called_function_name: optional parameter, if value is set, it needs to be included in returned code snippet.
    :return: source code snippet
    """
    source_code = open(source_code_files[file_id])
    source_code_lines = source_code.readlines()
    offset = -1
    function_call_string = source_code_lines[line_number + offset]
    if ")" in function_call_string and "(" in function_call_string:
        if function_call_string.index(")") < function_call_string.index("("):
            function_call_string = function_call_string[function_call_string.index(")") + 1 :]
    if ")" in function_call_string and "(" not in function_call_string:
        function_call_string = function_call_string[function_call_string.index(")") + 1 :]

    def __get_word_prior_to_bracket(string):
        if "(" not in string:
            return None
        string = string[: string.index("(")]
        string = string.split(" ")
        string = [e for e in string if len(e) > 0]
        string = string[-1]
        return string

    called_function_name_contained = called_function_name is None

    while (
        function_call_string.count("(") > function_call_string.count(")")
        or function_call_string.count("(") < 1
        or function_call_string.count(")") < 1
        or __get_word_prior_to_bracket(function_call_string) == "while"
        or __get_word_prior_to_bracket(function_call_string) == "for"
        or __get_word_prior_to_bracket(function_call_string) == "if"
        or not called_function_name_contained
    ):
        # if ) prior to (, cut first part away
        if ")" in function_call_string and "(" in function_call_string:
            if function_call_string.index(")") < function_call_string.index("("):
                function_call_string = function_call_string[function_call_string.index(")") + 1 :]
        # if word prior to ( is "while", "for" or "if", cut away until (
        word_prior_to_bracket = __get_word_prior_to_bracket(function_call_string)
        if word_prior_to_bracket is not None:
            if word_prior_to_bracket == "while" or word_prior_to_bracket == "for" or word_prior_to_bracket == "if":
                function_call_string = function_call_string[function_call_string.index("(") + 1 :]
        # check if called_function_name is contained in function_call_string
        if not called_function_name_contained:
            called_function_name_str = cast(str, called_function_name)
            if called_function_name_str in function_call_string:
                called_function_name_contained = True
        offset += 1
        function_call_string += source_code_lines[line_number + offset]
    function_call_string = function_call_string.replace("\n", "")
    # if called_function_name is set and contained more than once in function_call_string, split function_call_string
    if called_function_name is not None:
        while function_call_string.count(called_function_name) > 1:
            function_call_string = function_call_string[: function_call_string.rfind(called_function_name)]

    return function_call_string


def get_called_function_and_parameter_names_from_function_call(
    source_code_line: str, mangled_function_name: str, node: CUNode
) -> Tuple[Optional[str], List[Optional[str]]]:
    """Returns the name of the called function and the names of the variables used as parameters in a list,
    if any are used.
    If parameter is a complex expression (e.g. addition, or function call, None is used at the respective position.
    Returns None if function name not in source_code_line
    :param source_code_line: string, source code snippet which contains the function call
    :param mangled_function_name: mangled name of the called function
    :param node: CUNode
    :return: function and parameter names. None instead of specific parameter, if no respective variable could be found.
    """
    global __global_llvm_cxxfilt_path
    # find function name by finding biggest match between function call line and recursive call
    try:
        mangled_function_name = mangled_function_name.split(" ")[0]  # ignore line if present
        function_name = demangle(mangled_function_name).split("(")[0]
    except ValueError:
        return None, []
    except AttributeError:
        return None, []
    if function_name not in source_code_line:
        return None, []

    # get parameters in brackets
    # parameter_string = source_code_line[function_position:]
    parameter_string = source_code_line[source_code_line.find(function_name) + len(function_name) :]
    parameter_string = parameter_string.replace("\t", "")
    if ";" in parameter_string:
        parameter_string = parameter_string[: parameter_string.index(";")]
    # prune left
    while "(" in parameter_string and not parameter_string.startswith("("):
        parameter_string = parameter_string[parameter_string.find("(") :]
    # prune right
    while ")" in parameter_string and not parameter_string.endswith(")"):
        parameter_string = parameter_string[: parameter_string.rfind(")") + 1]
    # prune to correct amount of closing brackets
    while not parameter_string.count("(") == parameter_string.count(")"):
        parameter_string = parameter_string[:-1]
        parameter_string = parameter_string[: parameter_string.rfind(")") + 1]
    parameter_string = parameter_string[1:-1]
    # intersect parameters with set of known variables to prevent errors
    parameters = parameter_string.split(",")
    result_parameters: List[Optional[str]] = []
    for param in parameters:
        param = param.replace("\t", "")
        if "+" in param or "-" in param or "*" in param or "/" in param or "(" in param or ")" in param:
            param_expression = param.replace("+", "$$").replace("-", "$$").replace("/", "$$")
            param_expression = param_expression.replace("*", "$$").replace("(", "$$").replace(")", "$$")
            split_param_expression = param_expression.split("$$")
            split_param_expression = [ex.replace(" ", "") for ex in split_param_expression]
            # check if any of the parameters is in list of known variables
            split_param_expression = [
                ex
                for ex in split_param_expression
                if ex in [var.replace(".addr", "") for var in [v.name for v in node.local_vars + node.global_vars]]
            ]
            # check if type of any of them contains * (i.e. is a pointer)
            found_entry = False
            for var_name_to_check in split_param_expression:
                if found_entry:
                    break
                for known_var in node.local_vars + node.global_vars:
                    if found_entry:
                        break
                    if known_var.name.replace(".addr", "") == var_name_to_check:
                        if "*" in known_var.type:
                            result_parameters.append(var_name_to_check)
                            found_entry = True
            if not found_entry:
                result_parameters.append(None)
        else:
            # check if param in known variables:
            result_parameters.append(param.replace(" ", ""))
    return cast(Optional[str], function_name), result_parameters


def set_global_llvm_cxxfilt_path(value: str):
    """setter for __global_llvm_cxxfilt_path
    :param value: value to place in __global_llvm_cxxfilt_path"""
    global __global_llvm_cxxfilt_path
    __global_llvm_cxxfilt_path = value


def get_called_functions_recursively(
    pet: PEGraphX, root: Node, visited: List[Node], cache: Dict[Any, Any]
) -> List[Union[FunctionNode, DummyNode]]:
    """returns a recursively generated list of called functions, started at root."""
    visited.append(root)
    called_functions: List[Union[FunctionNode, DummyNode]] = []
    for child in [pet.node_at(cuid) for cuid in [e[1] for e in pet.out_edges(root.id)]]:
        # check if type is Func or Dummy
        if isinstance(child, (FunctionNode, DummyNode)):
            # CU contains a function call
            # if Dummy, map to Func
            if isinstance(child, DummyNode):
                for function_cu in pet.all_nodes(FunctionNode):
                    if child.name == function_cu.name:
                        child = function_cu
            called_functions.append(cast(Union[FunctionNode, DummyNode], child))
        elif child not in visited:
            if child in cache:
                called_functions += cache[child]
            else:
                # recursion step
                called_functions += get_called_functions_recursively(pet, child, visited, cache)
    if root not in cache:
        cache[root] = called_functions
    return called_functions


def contains_reduction(pet: PEGraphX, node: Node) -> bool:
    """Checks if node contains a reduction operation.
    :param pet: PET Graph
    :param node: CUNode
    :return: bool"""
    for red_var in pet.reduction_vars:
        if line_contained_in_region(red_var["reduction_line"], node.start_position(), node.end_position()):
            return True
    return False


def detect_mw_types(pet: PEGraphX, main_node: Node):
    """The mainNode we want to compute the Task Parallelism Pattern for it
    use Breadth First Search (BFS) to detect all barriers and workers.
    1.) all child nodes become first worker if they are not marked as worker before
    2.) if a child has dependence to more than one parent node, it will be marked as barrier
    Returns list of BARRIER_WORKER pairs 2
    :param pet: PET graph
    :param main_node: root node
    """

    # first insert all the direct children of main node in a queue to use it for the BFS
    for node in pet.direct_children_or_called_nodes(main_node):
        # a child node can be set to NONE or ROOT due a former detectMWNode call where it was the mainNode
        if node.mw_type == MWType.NONE or node.mw_type == MWType.ROOT:
            node.mw_type = MWType.FORK

        # while using the node as the base child, we copy all the other children in a copy vector.
        # we do that because it could be possible that two children of the current node (two dependency)
        # point to two different children of another child node which results that the child node becomes BARRIER
        # instead of WORKER
        # so we copy the whole other children in another vector and when one of the children of the current node
        # does point to the other child node, we just adjust mw_type and then we remove the node from the vector
        # Thus we prevent changing to BARRIER due of two dependencies pointing to two different children of
        # the other node

        # create the copy vector so that it only contains the other nodes
        other_nodes = pet.direct_children_or_called_nodes(main_node)
        other_nodes.remove(node)

        for other_node in other_nodes:
            if depends(pet, other_node, node):
                if other_node.mw_type == MWType.WORKER:
                    other_node.mw_type = MWType.BARRIER
                else:
                    other_node.mw_type = MWType.WORKER

                    # check if other_node has > 1 RAW dependencies to node
                    # -> not detected in previous step, since other_node is only
                    #    dependent of a single CU
                    raw_targets = []
                    for s, t, d in pet.out_edges(other_node.id):
                        if pet.node_at(t) == node:
                            if d.dtype == DepType.RAW:
                                raw_targets.append(t)
                    # remove entries which occur less than two times
                    raw_targets = [t for t in raw_targets if raw_targets.count(t) > 1]
                    # remove duplicates from list
                    raw_targets = list(set(raw_targets))
                    # if elements remaining, mark other_node as BARRIER
                    if len(raw_targets) > 0:
                        other_node.mw_type = MWType.BARRIER

    pairs = []
    # check for Barrier Worker pairs
    # if two barriers don't have any dependency to each other then they create a barrierWorker pair
    # so check every barrier pair that they don't have a dependency to each other -> barrierWorker
    direct_subnodes = pet.direct_children_or_called_nodes(main_node)
    for n1 in direct_subnodes:
        if n1.mw_type == MWType.BARRIER:
            for n2 in direct_subnodes:
                if n2.mw_type == MWType.BARRIER and n1 != n2:
                    if n2 in [pet.node_at(t) for s, t, d in pet.out_edges(n1.id)] or n2 in [
                        pet.node_at(s) for s, t, d in pet.in_edges(n1.id)
                    ]:
                        break
                    # so these two nodes are BarrierWorker, because there is no dependency between them
                    pairs.append((n1, n2))
                    n1.mw_type = MWType.BARRIER_WORKER
                    n2.mw_type = MWType.BARRIER_WORKER
    # return pairs


def get_var_definition_line_dict(cu_xml: str) -> Dict[str, List[str]]:
    """creates a dictionary {varname: [definitionLines]} based on cu_xml
    and return the dictionary.
    Removes .addr suffix if present.
    :param cu_xml: Path (string) to the CU xml file to be used
    :return: dictionary, containing information on variable definition lines
    """
    xml_fd = open(cu_xml)
    xml_content = ""
    for line in xml_fd.readlines():
        if not (line.rstrip().endswith("</Nodes>") or line.rstrip().endswith("<Nodes>")):
            xml_content = xml_content + line
    xml_content = "<Nodes>{0}</Nodes>".format(xml_content)
    parsed_cu = objectify.fromstring(xml_content)

    var_def_line_dict = dict()
    for node in parsed_cu.Node:
        # only consider cu nodes
        if node.get("type") == "0":
            # add global variables
            for idx, global_variables_entry in enumerate(node.globalVariables):
                try:
                    for i in global_variables_entry["global"]:
                        # insert mapping into var_def_line_dict
                        if not i.text.replace(".addr", "") in var_def_line_dict:
                            var_def_line_dict[i.text.replace(".addr", "")] = [i.get("defLine")]
                        else:
                            var_def_line_dict[i.text.replace(".addr", "")].append(i.get("defLine"))
                            var_def_line_dict[i.text.replace(".addr", "")] = list(
                                set(var_def_line_dict[i.text.replace(".addr", "")])
                            )
                except AttributeError:
                    pass
            # add local variables
            for idx, global_variables_entry in enumerate(node.localVariables):
                try:
                    for i in global_variables_entry["local"]:
                        # insert mapping into var_def_line_dict
                        if not i.text.replace(".addr", "") in var_def_line_dict:
                            var_def_line_dict[i.text.replace(".addr", "")] = [i.get("defLine")]
                        else:
                            var_def_line_dict[i.text.replace(".addr", "")].append(i.get("defLine"))
                            var_def_line_dict[i.text.replace(".addr", "")] = list(
                                set(var_def_line_dict[i.text.replace(".addr", "")])
                            )
                except AttributeError:
                    pass
    return var_def_line_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional, List, Set

from sympy import Symbol, Function, Integer  # type: ignore

from discopop_explorer.PEGraphX import NodeID
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID
from discopop_library.discopop_optimizer.classes.types.DataAccessType import (
    ReadDataAccess,
    WriteDataAccess,
)


class GenericNode(object):
    node_id: int  # id of the node in the nx.DiGraph which stores this object  # environment object to retrieve and store free symbols and other configurable values
    ## Performance modelling
    introduced_symbols: List[Symbol]
    performance_model: CostModel
    ## Data transfer calculation
    written_memory_regions: Set[WriteDataAccess]
    read_memory_regions: Set[ReadDataAccess]
    device_id: DeviceID
    execute_in_parallel: bool

    branch_affiliation: List[int]

    def __init__(
        self,
        node_id: int,
        experiment,
        cu_id: Optional[NodeID] = None,
        written_memory_regions: Optional[Set[WriteDataAccess]] = None,
        read_memory_regions: Optional[Set[ReadDataAccess]] = None,
        device_id: DeviceID = None,
    ):
        self.node_id = node_id
        self.cu_id = cu_id  # used to differentiate between "legacy" and suggestion nodes
        self.original_cu_id = cu_id  # used for the creation of update suggestions
        self.introduced_symbols = []
        self.performance_model = CostModel(Integer(0), Integer(0))
        self.suggestion = None
        self.suggestion_type: Optional[str] = None
        self.branch_affiliation = []
        self.execute_in_parallel = False

        if written_memory_regions is None:
            self.written_memory_regions = set()
        else:
            self.written_memory_regions = written_memory_regions
            # register free variables for memory region sizes
            for memory_region in [wmr.memory_region for wmr in written_memory_regions]:
                experiment.get_memory_region_size(memory_region, use_symbolic_value=True)

        if read_memory_regions is None:
            self.read_memory_regions = set()
        else:
            self.read_memory_regions = read_memory_regions
            # register free variables for memory region sizes
            for memory_region in [rmr.memory_region for rmr in read_memory_regions]:
                experiment.get_memory_region_size(memory_region, use_symbolic_value=True)

        self.device_id = device_id

    def get_plot_label(self) -> str:
        return ""

    def get_hover_text(self) -> str:
        return ""

    def get_cost_model(self, experiment, all_function_nodes, current_device) -> CostModel:
        raise NotImplementedError("Implementation needs to be provided by derived class: !", type(self))

    def register_child(self, other, experiment, all_function_nodes, current_device):
        """Registers a child node for the given model.
        Does not modify the stored model in self or other."""
        raise NotImplementedError("Implementation needs to be provided by derived class: !", type(self))

    def register_successor(self, other, root_node):
        """Registers a successor node for the given model.
        Does not modify the stored model in self or other."""
        raise NotImplementedError("Implementation needs to be provided by derived class: !", type(self))

    def represents_sequential_version(self) -> bool:
        """Returns True if the given node represents a sequential execution.
        Returns False, if the node characterizes a parallel option instead."""
        return self.suggestion is None
